import json
from typing import Dict, Union
from mattermostdriver import Driver
from sqlalchemy import create_engine
from mnemosynecore.vault.client import (
    get_connection_as_json,
    get_connection_as_json_test,
)


def _build_connection(cfg: Dict, conn_type: str) -> Union[Driver, "Engine", Dict, object]:
    conn_type = conn_type.lower()
    login = cfg.get("user") or cfg.get("login")
    extra = cfg.get("extra", {})
    if isinstance(extra, str):
        try:
            extra = json.loads(extra or "{}")
        except Exception:
            extra = {}
    if not isinstance(extra, dict):
        extra = {}

    if conn_type in ("vertica", "vertica_engine"):
        if not login:
            raise ValueError("Для Vertica нужен login или user")
        vertica_url = (
            f"vertica+vertica_python://{login}:{cfg['password']}@"
            f"{cfg['host']}:{cfg['port']}/{cfg.get('schema')}"
        )
        return create_engine(vertica_url, pool_pre_ping=True)

    elif conn_type == "clickhouse":
        from clickhouse_driver import Client
        if not login:
            raise ValueError("Для ClickHouse нужен login или user")

        return Client(
            host=cfg["host"],
            port=int(cfg.get("port", 9000)),
            user=login,
            password=cfg.get("password"),
            database=cfg.get("schema") or cfg.get("database") or extra.get("database"),
            secure=cfg.get("secure", extra.get("secure", False)),
        )

    elif conn_type == "superset":
        return {
            "host": cfg["host"],
            "login": login,
            "password": cfg.get("password"),
            "schema": cfg.get("schema", "https"),
            "port": int(cfg.get("port", 443)),
            "extra": extra,
        }

    elif conn_type == "mattermost":
        driver = Driver({
            "url": cfg["host"],
            "token": cfg["password"],
            "scheme": cfg.get("schema", "https"),
            "port": int(cfg.get("port", 443)),
            "basepath": extra.get("basepath", "/api/v4"),
        })
        driver.login()
        return driver

    elif conn_type == "raw":
        return cfg

    else:
        raise ValueError(f"Неизвестный conn_type: {conn_type}")


def un_conn(conn_id: str, conn_type: str) -> Union[Driver, "Engine", Dict, object]:
    cfg = json.loads(get_connection_as_json(conn_id))
    return _build_connection(cfg, conn_type)


def un_conn_test(
    conn_id: str,
    conn_type: str,
    dir_path: str | None = None,
    creds_path: str | None = None,
) -> Union[Driver, "Engine", Dict, object]:
    cfg = json.loads(
        get_connection_as_json_test(
            conn_id=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
    )
    return _build_connection(cfg, conn_type)



import json

import pytest

from mnemosynecore.vault import univ_conn


def _base_cfg(extra='{"basepath":"/custom"}'):
    return json.dumps(
        {
            "host": "localhost",
            "password": "pwd",
            "login": "usr",
            "port": 5433,
            "schema": "db",
            "extra": extra,
        }
    )


def test_un_conn_returns_engine_for_vertica(monkeypatch):
    captured = {}
    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg())
    monkeypatch.setattr(univ_conn, "create_engine", lambda url, pool_pre_ping=True: {"url": url, "pool_pre_ping": pool_pre_ping})
    engine = univ_conn.un_conn("CID", "vertica")
    assert "vertica+vertica_python://usr:pwd@localhost:5433/db" == engine["url"]
    assert engine["pool_pre_ping"] is True


def test_un_conn_supports_vertica_engine_alias(monkeypatch):
    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg())
    monkeypatch.setattr(univ_conn, "create_engine", lambda url, pool_pre_ping=True: "ENGINE")
    assert univ_conn.un_conn("CID", "vertica_engine") == "ENGINE"


def test_un_conn_clickhouse(monkeypatch):
    class FakeClient:
        def __init__(self, **kwargs):
            self.kwargs = kwargs

    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg(extra='{"secure": true}'))
    monkeypatch.setitem(__import__("sys").modules, "clickhouse_driver", type("M", (), {"Client": FakeClient})())
    client = univ_conn.un_conn("CID", "clickhouse")
    assert client.kwargs["database"] == "db"
    assert client.kwargs["secure"] is True


def test_un_conn_superset_parses_extra(monkeypatch):
    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg(extra='{"client_secret":"x"}'))
    cfg = univ_conn.un_conn("CID", "superset")
    assert cfg["host"] == "localhost"
    assert cfg["extra"]["client_secret"] == "x"


def test_un_conn_mattermost_uses_basepath(monkeypatch):
    class FakeDriver:
        def __init__(self, cfg):
            self.cfg = cfg
            self.logged_in = False

        def login(self):
            self.logged_in = True

    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg(extra='{"basepath":"/api/custom"}'))
    monkeypatch.setattr(univ_conn, "Driver", FakeDriver)
    driver = univ_conn.un_conn("CID", "mattermost")
    assert driver.cfg["basepath"] == "/api/custom"
    assert driver.logged_in is True


def test_un_conn_raw(monkeypatch):
    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg())
    cfg = univ_conn.un_conn("CID", "raw")
    assert cfg["login"] == "usr"


def test_un_conn_unknown_type(monkeypatch):
    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: _base_cfg())
    with pytest.raises(ValueError, match="Неизвестный conn_type"):
        univ_conn.un_conn("CID", "oops")


def test_un_conn_test_uses_test_secret(monkeypatch):
    monkeypatch.setattr(univ_conn, "get_connection_as_json_test", lambda *args, **kwargs: _base_cfg())
    monkeypatch.setattr(univ_conn, "create_engine", lambda url, pool_pre_ping=True: {"url": url})
    engine = univ_conn.un_conn_test("CID", "vertica", creds_path="/tmp/cid.json")
    assert "vertica+vertica_python://" in engine["url"]


def test_un_conn_vertica_supports_user_field(monkeypatch):
    payload = json.dumps(
        {
            "host": "localhost",
            "password": "pwd",
            "user": "usr_user",
            "port": 5433,
            "schema": "db",
            "extra": "{}",
        }
    )
    monkeypatch.setattr(univ_conn, "get_connection_as_json", lambda _: payload)
    monkeypatch.setattr(univ_conn, "create_engine", lambda url, pool_pre_ping=True: {"url": url})
    engine = univ_conn.un_conn("CID", "vertica")
    assert "usr_user:pwd@localhost:5433/db" in engine["url"]



import inspect
import json
import logging
import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urljoin

import requests

from mnemosynecore.vault.client import get_secret, get_secret_test

SizeValue = Union[str, Tuple[int, int], List[int]]
SupersetConn = Union[str, Dict[str, Any]]


def _parse_extra(extra_raw: Any) -> Dict[str, Any]:
    if isinstance(extra_raw, dict):
        return extra_raw
    if isinstance(extra_raw, str):
        try:
            parsed = json.loads(extra_raw) if extra_raw else {}
            if isinstance(parsed, dict):
                return parsed
        except json.JSONDecodeError:
            return {}
    return {}


def _resolve_superset_config(
    superset_conn: SupersetConn,
    dir_path: Optional[str] = None,
) -> Dict[str, Any]:
    if isinstance(superset_conn, dict):
        cfg = dict(superset_conn)
    elif isinstance(superset_conn, str):
        raw = superset_conn.strip()
        if raw.startswith("{"):
            try:
                cfg = json.loads(raw)
            except json.JSONDecodeError as exc:
                raise ValueError("Неверный формат JSON для конфигурации Superset") from exc
        else:
            from mnemosynecore.secrets import resolve_secret

            cfg = resolve_secret(superset_conn, dir_path)
    else:
        raise TypeError("superset_conn должен быть dict или str")

    extra = _parse_extra(cfg.get("extra"))
    out = dict(cfg)
    out["extra"] = extra
    out["host"] = out.get("host") or out.get("ss_url")
    out["ss_url"] = out.get("ss_url") or out.get("host")
    out["username"] = out.get("username") or out.get("login")
    out["login"] = out.get("login") or out.get("username")
    out["auth_provider"] = out.get("auth_provider") or extra.get("auth_provider") or "ldap"
    if out.get("security_cookie_auth_token") is None:
        out["security_cookie_auth_token"] = extra.get("security_cookie_auth_token")
    return out


def _resolve_superset_config_test(
    superset_conn: SupersetConn,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict[str, Any]:
    if isinstance(superset_conn, dict):
        return _resolve_superset_config(superset_conn, dir_path=None)
    if isinstance(superset_conn, str):
        raw = superset_conn.strip()
        if raw.startswith("{"):
            return _resolve_superset_config(raw, dir_path=None)
        return _resolve_superset_config(
            get_secret_test(raw, dir_path=dir_path, creds_path=creds_path),
            dir_path=None,
        )
    raise TypeError("superset_conn должен быть dict или str")


def _resolve_mattermost_bot_id_test(
    bot_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    raw = bot_id.strip()
    if raw.startswith("{"):
        try:
            json.loads(raw)
        except json.JSONDecodeError as exc:
            raise ValueError("Неверный формат JSON для конфигурации Mattermost") from exc
        return raw

    cfg = get_secret_test(raw, dir_path=dir_path, creds_path=creds_path)
    return json.dumps(cfg)


def _normalize_size_str(size: SizeValue) -> str:
    if isinstance(size, str):
        compact = size.replace(" ", "")
        if "," not in compact:
            raise ValueError(f"Некорректный размер: {size}")
        return compact
    if isinstance(size, (tuple, list)) and len(size) == 2:
        return f"{int(size[0])},{int(size[1])}"
    raise ValueError(f"Некорректный размер: {size}")


def _normalize_size_tuple(size: SizeValue) -> Tuple[int, int]:
    compact = _normalize_size_str(size)
    left, right = compact.split(",", 1)
    return int(left), int(right)


def _call_with_supported_kwargs(func, kwargs: Dict[str, Any]):
    try:
        return func(**kwargs)
    except TypeError:
        try:
            signature = inspect.signature(func)
        except (TypeError, ValueError):
            raise

        has_var_kwargs = any(
            p.kind == inspect.Parameter.VAR_KEYWORD
            for p in signature.parameters.values()
        )
        if has_var_kwargs:
            raise

        filtered = {}
        for name, param in signature.parameters.items():
            if param.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
            ) and name in kwargs:
                filtered[name] = kwargs[name]

        if not filtered or filtered == kwargs:
            raise
        return func(**filtered)


def _safe_name(value: str) -> str:
    sanitized = re.sub(r"[^A-Za-z0-9_.-]+", "_", value.strip())
    return sanitized.strip("_") or "screenshot"


def _normalize_superset_host(host: str) -> str:
    cleaned = host.strip()
    if not cleaned:
        raise ValueError("host Superset не может быть пустым")
    if cleaned.startswith("http://") or cleaned.startswith("https://"):
        return cleaned.rstrip("/")
    return f"https://{cleaned.rstrip('/')}"


def _build_superset_dashboard_url(
    *,
    cfg: Dict[str, Any],
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
) -> str:
    if dashboard_url:
        return dashboard_url
    if dashboard_id is None:
        raise ValueError("Нужно передать dashboard_id или dashboard_url")
    base = _normalize_superset_host(cfg["host"])
    return urljoin(base + "/", f"superset/dashboard/{dashboard_id}/")


def _build_superset_chart_url(
    *,
    cfg: Dict[str, Any],
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
) -> str:
    if chart_url:
        return chart_url
    if chart_id is None:
        raise ValueError("Нужно передать chart_id или chart_url")
    base = _normalize_superset_host(cfg["host"])
    return urljoin(base + "/", f"superset/explore/?slice_id={int(chart_id)}")


def _markdown_link(label: str, url: str) -> str:
    safe_label = (label or "").strip() or url
    return f"[{safe_label}]({url})"


def _get_mattermost_driver(bot_id: str):
    from mnemosynecore.mattermost import get_mattermost_driver

    return get_mattermost_driver(bot_id)


def get_superset_client(
    *,
    superset_conn: SupersetConn,
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
):
    cfg = _resolve_superset_config(superset_conn, dir_path)

    host = cfg.get("ss_url") or cfg.get("host")
    username = cfg.get("username") or cfg.get("login")
    password = cfg.get("password")
    auth_provider = cfg.get("auth_provider", "ldap")
    cookie_token = security_cookie_auth_token or cfg.get("security_cookie_auth_token")

    if not host:
        raise ValueError("Для Superset нужен host/ss_url")
    if not username:
        raise ValueError("Для Superset нужен login/username")
    if password is None:
        raise ValueError("Для Superset нужен password")

    try:
        from superset_o3_api_lib import superset_o3_api_lib

        superset_api = getattr(superset_o3_api_lib, "superset_api", None)
        if callable(superset_api):
            kwargs = {
                "username": username,
                "password": password,
                "ss_url": host,
                "auth_provider": auth_provider,
            }
            if cookie_token:
                kwargs["security_cookie_auth_token"] = cookie_token

            try:
                return superset_api(**kwargs)
            except TypeError:
                if cookie_token:
                    return superset_api(
                        username,
                        password,
                        host,
                        auth_provider,
                        security_cookie_auth_token=cookie_token,
                    )
                return superset_api(username, password, host, auth_provider)
    except ImportError:
        pass

    try:
        from superset_o3_api_lib import SupersetAPIClient
        from superset_o3_api_lib.auth import PasswordOAuthFlow
    except ImportError as exc:
        raise ImportError(
            "Для Superset-функций установите extra зависимость: mnemosynecore[superset]"
        ) from exc

    extra = cfg.get("extra") or {}
    client_secret = extra.get("client_secret")
    client_id = cfg.get("client_id") or cfg.get("schema") or extra.get("client_id")

    auth = PasswordOAuthFlow(
        superset_host=host,
        username=username,
        password=password,
        client_id=client_id,
        client_secret=client_secret,
    )
    return SupersetAPIClient(auth=auth)


def superset_request(
    *,
    endpoint: str,
    method: str = "GET",
    payload: dict | None = None,
    vault_conn_id: str,
):
    cfg = get_secret(vault_conn_id)
    return _superset_request_with_cfg(
        endpoint=endpoint,
        method=method,
        payload=payload,
        cfg=cfg,
    )


def _superset_request_with_cfg(
    *,
    endpoint: str,
    method: str = "GET",
    payload: dict | None = None,
    cfg: Dict[str, Any],
):
    base_url = cfg["host"]

    headers = {
        "Authorization": f"Bearer {cfg['password']}",
        "Content-Type": "application/json",
    }

    url = f"{base_url.rstrip('/')}/{endpoint.lstrip('/')}"
    resp = requests.request(method, url, json=payload, headers=headers)
    resp.raise_for_status()
    return resp.json()


def superset_request_test(
    *,
    endpoint: str,
    method: str = "GET",
    payload: dict | None = None,
    vault_conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    cfg = get_secret_test(vault_conn_id, dir_path=dir_path, creds_path=creds_path)
    return _superset_request_with_cfg(
        endpoint=endpoint,
        method=method,
        payload=payload,
        cfg=cfg,
    )


def superset_dashboard_url(
    *,
    superset_conn: SupersetConn,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    dir_path: Optional[str] = None,
) -> str:
    cfg = _resolve_superset_config(superset_conn, dir_path=dir_path)
    return _build_superset_dashboard_url(
        cfg=cfg,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
    )


def superset_chart_url(
    *,
    superset_conn: SupersetConn,
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
    dir_path: Optional[str] = None,
) -> str:
    cfg = _resolve_superset_config(superset_conn, dir_path=dir_path)
    return _build_superset_chart_url(
        cfg=cfg,
        chart_id=chart_id,
        chart_url=chart_url,
    )


def superset_dashboard_link(
    *,
    superset_conn: SupersetConn,
    dashboard_name: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    dir_path: Optional[str] = None,
) -> str:
    url = superset_dashboard_url(
        superset_conn=superset_conn,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
        dir_path=dir_path,
    )
    label = dashboard_name
    if not label:
        label = f"Dashboard {dashboard_id}" if dashboard_id is not None else "Dashboard"
    return _markdown_link(label, url)


def superset_chart_link(
    *,
    superset_conn: SupersetConn,
    chart_name: Optional[str] = None,
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
    dir_path: Optional[str] = None,
) -> str:
    url = superset_chart_url(
        superset_conn=superset_conn,
        chart_id=chart_id,
        chart_url=chart_url,
        dir_path=dir_path,
    )
    label = chart_name
    if not label:
        label = f"Chart {chart_id}" if chart_id is not None else "Chart"
    return _markdown_link(label, url)


def superset_dashboard_thumbnail(
    *,
    superset_conn: SupersetConn,
    dashboard_name: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> bytes:
    client = get_superset_client(
        superset_conn=superset_conn,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
    )

    if hasattr(client, "dashboard_thumbnail"):
        if not dashboard_name:
            raise ValueError(
                "dashboard_name обязателен для dashboard_thumbnail. "
                "Передайте имя дашборда или используйте старый API с dashboard_id/dashboard_url."
            )
        kwargs = {
            "dashboard_name": dashboard_name,
            "refresh": refresh,
            "refresh_wait_sec": refresh_wait_sec,
            "thumb_size": _normalize_size_str(thumb_size),
            "window_size": _normalize_size_str(window_size),
            "dashboard_url": dashboard_url,
        }
        return _call_with_supported_kwargs(client.dashboard_thumbnail, kwargs)

    if hasattr(client, "dashboard_screenshot"):
        if dashboard_id is None and dashboard_url is None:
            raise ValueError("Для dashboard_screenshot нужен dashboard_id или dashboard_url")
        kwargs = {
            "dashboard_id": dashboard_id,
            "dashboard_url": dashboard_url,
            "refresh": refresh,
            "refresh_wait_sec": refresh_wait_sec,
            "thumb_size": _normalize_size_tuple(thumb_size),
            "window_size": _normalize_size_tuple(window_size),
            "retry_count": 5,
        }
        return _call_with_supported_kwargs(client.dashboard_screenshot, kwargs)

    raise AttributeError(
        "Не найден метод dashboard_thumbnail/dashboard_screenshot в клиенте Superset API"
    )


def superset_chart_thumbnail(
    *,
    superset_conn: SupersetConn,
    chart_id: int,
    force_refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> bytes:
    client = get_superset_client(
        superset_conn=superset_conn,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
    )

    chart_method = None
    if hasattr(client, "chart_thumbnail"):
        chart_method = client.chart_thumbnail
    elif hasattr(client, "chart_screenshot"):
        chart_method = client.chart_screenshot

    if chart_method is None:
        raise AttributeError(
            "Не найден метод chart_thumbnail/chart_screenshot в клиенте Superset API"
        )

    kwargs = {
        "chart_id": chart_id,
        "force_refresh": force_refresh,
        "refresh_wait_sec": refresh_wait_sec,
        "thumb_size": _normalize_size_str(thumb_size),
        "window_size": _normalize_size_str(window_size),
    }
    return _call_with_supported_kwargs(chart_method, kwargs)


def superset_screenshot_dashboard(
    conn_id: SupersetConn,
    dashboards: List[Dict],
    dir_path: Optional[str] = None,
    output_dir: str = ".",
) -> List[str]:
    os.makedirs(output_dir, exist_ok=True)
    saved_files: List[str] = []

    for i, dashboard in enumerate(dashboards, start=1):
        screenshot = superset_dashboard_thumbnail(
            superset_conn=conn_id,
            dashboard_name=dashboard.get("dashboard_name") or dashboard.get("name"),
            dashboard_id=dashboard.get("id"),
            dashboard_url=dashboard.get("dashboard_url") or dashboard.get("url"),
            refresh=dashboard.get("refresh", True),
            refresh_wait_sec=dashboard.get("refresh_wait_sec", 60),
            thumb_size=dashboard.get("thumb_size", "2048,1536"),
            window_size=dashboard.get("window_size", "2048,1536"),
            dir_path=dir_path,
            security_cookie_auth_token=dashboard.get("security_cookie_auth_token"),
        )

        name = (
            dashboard.get("file_name")
            or dashboard.get("dashboard_name")
            or dashboard.get("name")
            or str(dashboard.get("id") or "dashboard")
        )
        filename = os.path.join(output_dir, f"{i}_{_safe_name(str(name))}.png")

        with open(filename, "wb") as f:
            f.write(screenshot)
        saved_files.append(filename)

    return saved_files


def superset_screenshot_charts(
    conn_id: SupersetConn,
    charts: List[Dict],
    dir_path: Optional[str] = None,
    output_dir: str = ".",
) -> List[str]:
    os.makedirs(output_dir, exist_ok=True)
    saved_files: List[str] = []

    for i, chart in enumerate(charts, start=1):
        chart_id = chart.get("chart_id") or chart.get("id")
        if chart_id is None:
            raise ValueError("В списке charts у каждого элемента должен быть chart_id (или id)")

        screenshot = superset_chart_thumbnail(
            superset_conn=conn_id,
            chart_id=int(chart_id),
            force_refresh=chart.get("force_refresh", True),
            refresh_wait_sec=chart.get("refresh_wait_sec", 60),
            thumb_size=chart.get("thumb_size", "2048,1536"),
            window_size=chart.get("window_size", "2048,1536"),
            dir_path=dir_path,
            security_cookie_auth_token=chart.get("security_cookie_auth_token"),
        )

        name = chart.get("file_name") or f"chart_{chart_id}"
        filename = os.path.join(output_dir, f"{i}_{_safe_name(str(name))}.png")

        with open(filename, "wb") as f:
            f.write(screenshot)
        saved_files.append(filename)

    return saved_files


def _post_png_to_mattermost(
    *,
    channel_id: str,
    bot_id: str,
    png_bytes: bytes,
    file_name: str,
    text: str = "",
    root_id: Optional[str] = None,
) -> Dict[str, Optional[str]]:
    driver = _get_mattermost_driver(bot_id)
    upload = driver.files.upload_file(
        channel_id=channel_id,
        files={"files": (file_name, png_bytes, "image/png")},
    )

    file_infos = upload.get("file_infos") if isinstance(upload, dict) else None
    if not file_infos:
        raise ValueError("Mattermost не вернул file_infos после загрузки файла")

    first_file = file_infos[0]
    file_id = first_file.get("id") if isinstance(first_file, dict) else str(first_file)
    message = text.strip() if text else " "

    post = driver.posts.create_post(
        options={
            "channel_id": channel_id,
            "message": message,
            "file_ids": [file_id],
            **({"root_id": root_id} if root_id else {}),
        }
    )

    post_id = post.get("id") if isinstance(post, dict) else None
    return {"file_id": file_id, "post_id": post_id, "root_id": root_id}


def send_superset_dashboard_screenshot(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboard_name: str,
    dashboard_url: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    text: str = "",
    file_name: Optional[str] = None,
    root_id: Optional[str] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    png_bytes = superset_dashboard_thumbnail(
        superset_conn=superset_conn,
        dashboard_name=dashboard_name,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
    )

    final_name = file_name or f"dashboard_{_safe_name(dashboard_name)}.png"
    result = _post_png_to_mattermost(
        channel_id=channel_id,
        bot_id=bot_id,
        png_bytes=png_bytes,
        file_name=final_name,
        text=text,
        root_id=root_id,
    )
    if not silent:
        logging.info(
            "Скриншот дашборда отправлен в Mattermost: channel_id=%s, file=%s",
            channel_id,
            final_name,
        )
    return {
        "channel_id": channel_id,
        "file_name": final_name,
        "file_id": result.get("file_id"),
        "post_id": result.get("post_id"),
        "root_id": result.get("root_id"),
    }


def send_superset_dashboard_permalink_screenshot(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboard_url: str,
    dashboard_name: str,
    text: str = "",
    include_link: bool = True,
    link_text: Optional[str] = None,
    file_name: Optional[str] = None,
    root_id: Optional[str] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    final_text = text
    if include_link:
        link = _markdown_link(link_text or dashboard_name, dashboard_url)
        final_text = _join_text_parts(text, link)

    return send_superset_dashboard_screenshot(
        channel_id=channel_id,
        bot_id=bot_id,
        superset_conn=superset_conn,
        dashboard_name=dashboard_name,
        dashboard_url=dashboard_url,
        text=final_text,
        file_name=file_name,
        root_id=root_id,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
        silent=silent,
    )


def send_superset_chart_screenshot(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    chart_id: int,
    text: str = "",
    file_name: Optional[str] = None,
    root_id: Optional[str] = None,
    force_refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    png_bytes = superset_chart_thumbnail(
        superset_conn=superset_conn,
        chart_id=chart_id,
        force_refresh=force_refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
    )

    final_name = file_name or f"chart_{chart_id}.png"
    result = _post_png_to_mattermost(
        channel_id=channel_id,
        bot_id=bot_id,
        png_bytes=png_bytes,
        file_name=final_name,
        text=text,
        root_id=root_id,
    )
    if not silent:
        logging.info(
            "Скриншот чарта отправлен в Mattermost: channel_id=%s, chart_id=%s",
            channel_id,
            chart_id,
        )
    return {
        "channel_id": channel_id,
        "file_name": final_name,
        "file_id": result.get("file_id"),
        "post_id": result.get("post_id"),
        "root_id": result.get("root_id"),
    }


def superset_dashboard_screenshot_to_file(
    *,
    superset_conn: SupersetConn,
    output_path: str,
    dashboard_name: Optional[str] = None,
    dashboard_url: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> str:
    png_bytes = superset_dashboard_thumbnail(
        superset_conn=superset_conn,
        dashboard_name=dashboard_name,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
    )

    output_dir = os.path.dirname(output_path)
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
    with open(output_path, "wb") as f:
        f.write(png_bytes)
    return output_path


def superset_chart_screenshot_to_file(
    *,
    superset_conn: SupersetConn,
    chart_id: int,
    output_path: str,
    force_refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> str:
    png_bytes = superset_chart_thumbnail(
        superset_conn=superset_conn,
        chart_id=chart_id,
        force_refresh=force_refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        dir_path=dir_path,
        security_cookie_auth_token=security_cookie_auth_token,
    )

    output_dir = os.path.dirname(output_path)
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
    with open(output_path, "wb") as f:
        f.write(png_bytes)
    return output_path


def _merge_text_with_link(text: str, link: str) -> str:
    clean_text = text.strip()
    if clean_text:
        return f"{clean_text}\n{link}"
    return link


def _join_text_parts(*parts: str) -> str:
    cleaned = [p.strip() for p in parts if p and p.strip()]
    return "\n".join(cleaned)


def send_superset_dashboard_link(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboard_name: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    text: str = "",
    root_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    from mnemosynecore.mattermost import post_message

    link = superset_dashboard_link(
        superset_conn=superset_conn,
        dashboard_name=dashboard_name,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
        dir_path=dir_path,
    )
    result = post_message(
        channel_id=channel_id,
        bot_id=bot_id,
        text=_merge_text_with_link(text, link),
        root_id=root_id,
        silent=silent,
    )
    result["link"] = link
    return result


def send_superset_chart_link(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    chart_name: Optional[str] = None,
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
    text: str = "",
    root_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    from mnemosynecore.mattermost import post_message

    link = superset_chart_link(
        superset_conn=superset_conn,
        chart_name=chart_name,
        chart_id=chart_id,
        chart_url=chart_url,
        dir_path=dir_path,
    )
    result = post_message(
        channel_id=channel_id,
        bot_id=bot_id,
        text=_merge_text_with_link(text, link),
        root_id=root_id,
        silent=silent,
    )
    result["link"] = link
    return result


def send_superset_dashboard_links_to_thread(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboards: List[Dict[str, Any]],
    header_text: str = "",
    root_id: Optional[str] = None,
    thread_messages: Optional[List[str]] = None,
    dir_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Any]:
    from mnemosynecore.mattermost import send_thread_messages

    if not dashboards:
        return {
            "main_post_id": None,
            "root_id": root_id,
            "links_sent": [],
            "thread_messages": {"sent": [], "failed": []},
        }

    links_sent: List[Dict[str, Optional[str]]] = []
    main_post_id = None
    active_root_id = root_id

    for index, dashboard in enumerate(dashboards):
        dashboard_name = dashboard.get("dashboard_name") or dashboard.get("name")
        dashboard_id = dashboard.get("dashboard_id") or dashboard.get("id")
        dashboard_url = dashboard.get("dashboard_url") or dashboard.get("url")
        text = dashboard.get("text", "")
        if index == 0:
            text = _join_text_parts(header_text, text)

        result = send_superset_dashboard_link(
            channel_id=channel_id,
            bot_id=bot_id,
            superset_conn=superset_conn,
            dashboard_name=dashboard_name,
            dashboard_id=dashboard_id,
            dashboard_url=dashboard_url,
            text=text,
            root_id=active_root_id if index > 0 or active_root_id else None,
            dir_path=dir_path,
            silent=silent,
        )
        links_sent.append(result)
        if index == 0 and not active_root_id:
            main_post_id = result.get("post_id")
            active_root_id = main_post_id

    thread_result = {"sent": [], "failed": []}
    if thread_messages and active_root_id:
        thread_result = send_thread_messages(
            channel_id=channel_id,
            bot_id=bot_id,
            root_id=active_root_id,
            messages=thread_messages,
            silent=silent,
            continue_on_error=True,
        )

    return {
        "main_post_id": main_post_id,
        "root_id": active_root_id,
        "links_sent": links_sent,
        "thread_messages": thread_result,
    }


def send_superset_chart_links_to_thread(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    charts: List[Dict[str, Any]],
    header_text: str = "",
    root_id: Optional[str] = None,
    thread_messages: Optional[List[str]] = None,
    dir_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Any]:
    from mnemosynecore.mattermost import send_thread_messages

    if not charts:
        return {
            "main_post_id": None,
            "root_id": root_id,
            "links_sent": [],
            "thread_messages": {"sent": [], "failed": []},
        }

    links_sent: List[Dict[str, Optional[str]]] = []
    main_post_id = None
    active_root_id = root_id

    for index, chart in enumerate(charts):
        chart_name = chart.get("chart_name") or chart.get("name")
        chart_id = chart.get("chart_id") or chart.get("id")
        chart_url = chart.get("chart_url") or chart.get("url")
        text = chart.get("text", "")
        if index == 0:
            text = _join_text_parts(header_text, text)

        result = send_superset_chart_link(
            channel_id=channel_id,
            bot_id=bot_id,
            superset_conn=superset_conn,
            chart_name=chart_name,
            chart_id=chart_id,
            chart_url=chart_url,
            text=text,
            root_id=active_root_id if index > 0 or active_root_id else None,
            dir_path=dir_path,
            silent=silent,
        )
        links_sent.append(result)
        if index == 0 and not active_root_id:
            main_post_id = result.get("post_id")
            active_root_id = main_post_id

    thread_result = {"sent": [], "failed": []}
    if thread_messages and active_root_id:
        thread_result = send_thread_messages(
            channel_id=channel_id,
            bot_id=bot_id,
            root_id=active_root_id,
            messages=thread_messages,
            silent=silent,
            continue_on_error=True,
        )

    return {
        "main_post_id": main_post_id,
        "root_id": active_root_id,
        "links_sent": links_sent,
        "thread_messages": thread_result,
    }


def send_superset_dashboards_to_channels(
    *,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboards: List[Dict[str, Any]],
    default_channel_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    silent: bool = False,
) -> List[Dict[str, Optional[str]]]:
    results: List[Dict[str, Optional[str]]] = []

    for dashboard in dashboards:
        channel_id = dashboard.get("channel_id") or default_channel_id
        if not channel_id:
            raise ValueError("Для каждого dashboard нужен channel_id или default_channel_id")

        dashboard_name = dashboard.get("dashboard_name") or dashboard.get("name")
        if not dashboard_name:
            raise ValueError("Для каждого dashboard нужен dashboard_name (или name)")

        result = send_superset_dashboard_screenshot(
            channel_id=channel_id,
            bot_id=bot_id,
            superset_conn=superset_conn,
            dashboard_name=dashboard_name,
            dashboard_url=dashboard.get("dashboard_url") or dashboard.get("url"),
            dashboard_id=dashboard.get("id"),
            text=dashboard.get("text", ""),
            file_name=dashboard.get("file_name"),
            root_id=dashboard.get("root_id"),
            refresh=dashboard.get("refresh", True),
            refresh_wait_sec=dashboard.get("refresh_wait_sec", 60),
            thumb_size=dashboard.get("thumb_size", "2048,1536"),
            window_size=dashboard.get("window_size", "2048,1536"),
            dir_path=dir_path,
            security_cookie_auth_token=dashboard.get("security_cookie_auth_token"),
            silent=silent,
        )
        results.append(result)

    return results


def send_superset_charts_to_channels(
    *,
    bot_id: str,
    superset_conn: SupersetConn,
    charts: List[Dict[str, Any]],
    default_channel_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    silent: bool = False,
) -> List[Dict[str, Optional[str]]]:
    results: List[Dict[str, Optional[str]]] = []

    for chart in charts:
        channel_id = chart.get("channel_id") or default_channel_id
        if not channel_id:
            raise ValueError("Для каждого chart нужен channel_id или default_channel_id")

        chart_id = chart.get("chart_id") or chart.get("id")
        if chart_id is None:
            raise ValueError("Для каждого chart нужен chart_id (или id)")

        result = send_superset_chart_screenshot(
            channel_id=channel_id,
            bot_id=bot_id,
            superset_conn=superset_conn,
            chart_id=int(chart_id),
            text=chart.get("text", ""),
            file_name=chart.get("file_name"),
            root_id=chart.get("root_id"),
            force_refresh=chart.get("force_refresh", True),
            refresh_wait_sec=chart.get("refresh_wait_sec", 60),
            thumb_size=chart.get("thumb_size", "2048,1536"),
            window_size=chart.get("window_size", "2048,1536"),
            dir_path=dir_path,
            security_cookie_auth_token=chart.get("security_cookie_auth_token"),
            silent=silent,
        )
        results.append(result)

    return results


def get_superset_client_test(
    *,
    superset_conn: SupersetConn,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
):
    return get_superset_client(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        security_cookie_auth_token=security_cookie_auth_token,
    )


def superset_dashboard_thumbnail_test(
    *,
    superset_conn: SupersetConn,
    dashboard_name: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> bytes:
    return superset_dashboard_thumbnail(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboard_name=dashboard_name,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
    )


def superset_chart_thumbnail_test(
    *,
    superset_conn: SupersetConn,
    chart_id: int,
    force_refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> bytes:
    return superset_chart_thumbnail(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        chart_id=chart_id,
        force_refresh=force_refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
    )


def superset_screenshot_dashboard_test(
    conn_id: SupersetConn,
    dashboards: List[Dict],
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    output_dir: str = ".",
) -> List[str]:
    return superset_screenshot_dashboard(
        conn_id=_resolve_superset_config_test(
            superset_conn=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboards=dashboards,
        output_dir=output_dir,
    )


def superset_screenshot_charts_test(
    conn_id: SupersetConn,
    charts: List[Dict],
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    output_dir: str = ".",
) -> List[str]:
    return superset_screenshot_charts(
        conn_id=_resolve_superset_config_test(
            superset_conn=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        charts=charts,
        output_dir=output_dir,
    )


def send_superset_dashboard_screenshot_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboard_name: str,
    dashboard_url: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    text: str = "",
    file_name: Optional[str] = None,
    root_id: Optional[str] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_superset_dashboard_screenshot(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboard_name=dashboard_name,
        dashboard_url=dashboard_url,
        dashboard_id=dashboard_id,
        text=text,
        file_name=file_name,
        root_id=root_id,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
        silent=silent,
    )


def send_superset_chart_screenshot_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    chart_id: int,
    text: str = "",
    file_name: Optional[str] = None,
    root_id: Optional[str] = None,
    force_refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_superset_chart_screenshot(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        chart_id=chart_id,
        text=text,
        file_name=file_name,
        root_id=root_id,
        force_refresh=force_refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
        silent=silent,
    )


def send_superset_dashboard_permalink_screenshot_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboard_url: str,
    dashboard_name: str,
    text: str = "",
    include_link: bool = True,
    link_text: Optional[str] = None,
    file_name: Optional[str] = None,
    root_id: Optional[str] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_superset_dashboard_permalink_screenshot(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboard_url=dashboard_url,
        dashboard_name=dashboard_name,
        text=text,
        include_link=include_link,
        link_text=link_text,
        file_name=file_name,
        root_id=root_id,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
        silent=silent,
    )


def send_superset_dashboards_to_channels_test(
    *,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboards: List[Dict[str, Any]],
    default_channel_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    silent: bool = False,
) -> List[Dict[str, Optional[str]]]:
    return send_superset_dashboards_to_channels(
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboards=dashboards,
        default_channel_id=default_channel_id,
        silent=silent,
    )


def send_superset_charts_to_channels_test(
    *,
    bot_id: str,
    superset_conn: SupersetConn,
    charts: List[Dict[str, Any]],
    default_channel_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    silent: bool = False,
) -> List[Dict[str, Optional[str]]]:
    return send_superset_charts_to_channels(
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        charts=charts,
        default_channel_id=default_channel_id,
        silent=silent,
    )


def superset_dashboard_url_test(
    *,
    superset_conn: SupersetConn,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    return superset_dashboard_url(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
    )


def superset_chart_url_test(
    *,
    superset_conn: SupersetConn,
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    return superset_chart_url(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        chart_id=chart_id,
        chart_url=chart_url,
    )


def superset_dashboard_link_test(
    *,
    superset_conn: SupersetConn,
    dashboard_name: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    return superset_dashboard_link(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboard_name=dashboard_name,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
    )


def superset_chart_link_test(
    *,
    superset_conn: SupersetConn,
    chart_name: Optional[str] = None,
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    return superset_chart_link(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        chart_name=chart_name,
        chart_id=chart_id,
        chart_url=chart_url,
    )


def superset_dashboard_screenshot_to_file_test(
    *,
    superset_conn: SupersetConn,
    output_path: str,
    dashboard_name: Optional[str] = None,
    dashboard_url: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> str:
    return superset_dashboard_screenshot_to_file(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        output_path=output_path,
        dashboard_name=dashboard_name,
        dashboard_url=dashboard_url,
        dashboard_id=dashboard_id,
        refresh=refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
    )


def superset_chart_screenshot_to_file_test(
    *,
    superset_conn: SupersetConn,
    chart_id: int,
    output_path: str,
    force_refresh: Union[bool, str] = True,
    refresh_wait_sec: int = 60,
    thumb_size: SizeValue = "2048,1536",
    window_size: SizeValue = "2048,1536",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    security_cookie_auth_token: Optional[str] = None,
) -> str:
    return superset_chart_screenshot_to_file(
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        chart_id=chart_id,
        output_path=output_path,
        force_refresh=force_refresh,
        refresh_wait_sec=refresh_wait_sec,
        thumb_size=thumb_size,
        window_size=window_size,
        security_cookie_auth_token=security_cookie_auth_token,
    )


def send_superset_dashboard_link_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboard_name: Optional[str] = None,
    dashboard_id: Optional[int] = None,
    dashboard_url: Optional[str] = None,
    text: str = "",
    root_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_superset_dashboard_link(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboard_name=dashboard_name,
        dashboard_id=dashboard_id,
        dashboard_url=dashboard_url,
        text=text,
        root_id=root_id,
        silent=silent,
    )


def send_superset_chart_link_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    chart_name: Optional[str] = None,
    chart_id: Optional[int] = None,
    chart_url: Optional[str] = None,
    text: str = "",
    root_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_superset_chart_link(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        chart_name=chart_name,
        chart_id=chart_id,
        chart_url=chart_url,
        text=text,
        root_id=root_id,
        silent=silent,
    )


def send_superset_dashboard_links_to_thread_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    dashboards: List[Dict[str, Any]],
    header_text: str = "",
    root_id: Optional[str] = None,
    thread_messages: Optional[List[str]] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Any]:
    return send_superset_dashboard_links_to_thread(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        dashboards=dashboards,
        header_text=header_text,
        root_id=root_id,
        thread_messages=thread_messages,
        silent=silent,
    )


def send_superset_chart_links_to_thread_test(
    *,
    channel_id: str,
    bot_id: str,
    superset_conn: SupersetConn,
    charts: List[Dict[str, Any]],
    header_text: str = "",
    root_id: Optional[str] = None,
    thread_messages: Optional[List[str]] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Any]:
    return send_superset_chart_links_to_thread(
        channel_id=channel_id,
        bot_id=_resolve_mattermost_bot_id_test(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        superset_conn=_resolve_superset_config_test(
            superset_conn=superset_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        charts=charts,
        header_text=header_text,
        root_id=root_id,
        thread_messages=thread_messages,
        silent=silent,
    )



import json
import os
from typing import Any, Dict, Iterable, Optional


class VaultClient:
    def __init__(self):
        pass

    def get_secret(self, conn_id: str) -> Optional[str]:
        return None


def _airflow_connection_as_json(conn_name: str) -> str:
    try:
        from airflow.hooks.base import BaseHook
    except ImportError:
        from airflow.hooks.base_hook import BaseHook

    conn = BaseHook.get_connection(conn_name)
    ci = {
        "host": conn.host,
        "password": conn.password,
        "login": conn.login,
        "port": conn.port,
        "schema": conn.schema,
        "extra": conn.extra,
    }
    return json.dumps(ci)


def get_connection_as_json(conn_name: str) -> str:
    secret = os.environ.get(conn_name)
    if secret:
        return secret

    try:
        vault_client = VaultClient()
        secret = vault_client.get_secret(conn_name)
        if secret:
            return secret
    except Exception:
        pass

    try:
        return _airflow_connection_as_json(conn_name)
    except ImportError:
        pass
    except Exception:
        pass

    raise ValueError(
        f"Секрет {conn_name} не найден в переменных окружения, Vault или Airflow"
    )


def get_secret(conn_id: str) -> Dict:
    raw = get_connection_as_json(conn_id)
    try:
        return json.loads(raw)
    except json.JSONDecodeError as e:
        raise ValueError(f"Секрет {conn_id} не является корректным JSON: {e}")


def _get_test_secret_candidates(
    conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> list[str]:
    filename = f"{conn_id}.json"
    candidates: list[str] = []

    if creds_path:
        abs_creds_path = os.path.abspath(creds_path)
        if os.path.isdir(abs_creds_path):
            candidates.append(os.path.join(abs_creds_path, filename))
        else:
            candidates.append(abs_creds_path)

    if dir_path:
        abs_dir_path = os.path.abspath(dir_path)
        if os.path.isdir(abs_dir_path):
            candidates.append(os.path.join(abs_dir_path, filename))
        else:
            candidates.append(abs_dir_path)

    candidates.append(os.path.join(os.getcwd(), filename))
    return candidates


def get_connection_as_json_test(
    conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    candidates = _get_test_secret_candidates(
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )

    for file_path in candidates:
        if os.path.isfile(file_path):
            with open(file_path, "r", encoding="utf-8") as f:
                return f.read()

    raise FileNotFoundError(
        f"Тестовый секрет {conn_id}.json не найден. Проверены пути: {candidates}"
    )


def get_secret_test(
    conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict:
    if creds_path is None:
        raw = get_connection_as_json_test(conn_id, dir_path=dir_path)
    else:
        raw = get_connection_as_json_test(
            conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
    try:
        return json.loads(raw)
    except json.JSONDecodeError as e:
        raise ValueError(f"Тестовый секрет {conn_id} не является корректным JSON: {e}")


def _get_secret_test_compat(
    conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict:
    if creds_path is None:
        return get_secret_test(conn_id, dir_path=dir_path)
    return get_secret_test(conn_id, dir_path=dir_path, creds_path=creds_path)


def has_connection(conn_name: str) -> bool:
    try:
        get_connection_as_json(conn_name)
        return True
    except Exception:
        return False


def _deep_get(data: Dict[str, Any], field_path: str) -> Any:
    if not field_path:
        return data
    current: Any = data
    for part in field_path.split("."):
        if isinstance(current, dict) and part in current:
            current = current[part]
        else:
            raise KeyError(field_path)
    return current


def get_secret_field(
    conn_id: str,
    field_path: str,
    default: Any = None,
    *,
    required: bool = False,
    test: bool = False,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Any:
    secret = (
        _get_secret_test_compat(
            conn_id=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
        if test
        else get_secret(conn_id)
    )
    try:
        value = _deep_get(secret, field_path)
    except KeyError:
        if required:
            raise KeyError(f"Поле '{field_path}' отсутствует в секрете '{conn_id}'")
        return default

    if value is None and required:
        raise ValueError(f"Поле '{field_path}' в секрете '{conn_id}' пустое")
    return value


def require_secret_fields(
    conn_id: str,
    required_fields: Iterable[str],
    *,
    test: bool = False,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict[str, Any]:
    secret = (
        _get_secret_test_compat(
            conn_id=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
        if test
        else get_secret(conn_id)
    )
    missing = []

    for field in required_fields:
        try:
            value = _deep_get(secret, field)
        except KeyError:
            missing.append(field)
            continue
        if value is None or (isinstance(value, str) and not value.strip()):
            missing.append(field)

    if missing:
        raise ValueError(
            f"В секрете '{conn_id}' отсутствуют обязательные поля: {', '.join(missing)}"
        )
    return secret


def get_secret_with_defaults(
    conn_id: str,
    defaults: Dict[str, Any],
    *,
    test: bool = False,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict[str, Any]:
    secret = (
        _get_secret_test_compat(
            conn_id=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
        if test
        else get_secret(conn_id)
    )
    out = dict(defaults)
    out.update(secret)
    return out



from pathlib import Path

import pytest

from mnemosynecore.vault import client


def test_get_connection_as_json_from_env(monkeypatch):
    payload = '{"host":"env"}'
    monkeypatch.setenv("MY_CONN", payload)
    assert client.get_connection_as_json("MY_CONN") == payload


def test_get_connection_as_json_from_vault(monkeypatch):
    monkeypatch.delenv("MY_CONN", raising=False)
    monkeypatch.setattr(client.VaultClient, "get_secret", lambda self, key: '{"host":"vault"}')
    monkeypatch.setattr(client, "_airflow_connection_as_json", lambda key: '{"host":"airflow"}')
    assert client.get_connection_as_json("MY_CONN") == '{"host":"vault"}'


def test_get_connection_as_json_from_airflow(monkeypatch):
    monkeypatch.delenv("MY_CONN", raising=False)
    monkeypatch.setattr(client.VaultClient, "get_secret", lambda self, key: None)
    monkeypatch.setattr(client, "_airflow_connection_as_json", lambda key: '{"host":"airflow"}')
    assert client.get_connection_as_json("MY_CONN") == '{"host":"airflow"}'


def test_get_connection_as_json_raises_when_missing(monkeypatch):
    monkeypatch.delenv("MY_CONN", raising=False)
    monkeypatch.setattr(client.VaultClient, "get_secret", lambda self, key: None)
    monkeypatch.setattr(client, "_airflow_connection_as_json", lambda key: (_ for _ in ()).throw(ImportError()))
    with pytest.raises(ValueError, match="MY_CONN"):
        client.get_connection_as_json("MY_CONN")


def test_get_secret_parses_json(monkeypatch):
    monkeypatch.setattr(client, "get_connection_as_json", lambda _: '{"a": 1}')
    assert client.get_secret("X") == {"a": 1}


def test_get_secret_raises_on_invalid_json(monkeypatch):
    monkeypatch.setattr(client, "get_connection_as_json", lambda _: "not-json")
    with pytest.raises(ValueError, match="не является корректным JSON"):
        client.get_secret("X")


def test_get_connection_as_json_test_from_dir_path(tmp_path: Path):
    file = tmp_path / "BOT.json"
    file.write_text('{"x": 1}', encoding="utf-8")
    assert client.get_connection_as_json_test("BOT", dir_path=str(tmp_path)) == '{"x": 1}'


def test_get_connection_as_json_test_from_explicit_file_path(tmp_path: Path):
    file = tmp_path / "my_bot_creds.json"
    file.write_text('{"x": 3}', encoding="utf-8")
    assert client.get_connection_as_json_test("BOT", creds_path=str(file)) == '{"x": 3}'


def test_get_connection_as_json_test_from_cwd(monkeypatch, tmp_path: Path):
    file = tmp_path / "BOT.json"
    file.write_text('{"x": 2}', encoding="utf-8")
    monkeypatch.chdir(tmp_path)
    assert client.get_connection_as_json_test("BOT") == '{"x": 2}'


def test_get_connection_as_json_test_not_found(tmp_path: Path):
    with pytest.raises(FileNotFoundError, match="BOT.json"):
        client.get_connection_as_json_test("BOT", dir_path=str(tmp_path))


def test_get_secret_test_parses_json(monkeypatch):
    monkeypatch.setattr(client, "get_connection_as_json_test", lambda conn_id, dir_path=None: '{"a": 2}')
    assert client.get_secret_test("X") == {"a": 2}


def test_get_secret_test_parses_json_with_creds_path(monkeypatch):
    monkeypatch.setattr(
        client,
        "get_connection_as_json_test",
        lambda conn_id, dir_path=None, creds_path=None: '{"a": 5}',
    )
    assert client.get_secret_test("X", creds_path="/tmp/x.json") == {"a": 5}


def test_get_secret_test_raises_on_invalid_json(monkeypatch):
    monkeypatch.setattr(client, "get_connection_as_json_test", lambda conn_id, dir_path=None: "bad")
    with pytest.raises(ValueError, match="не является корректным JSON"):
        client.get_secret_test("X")


def test_has_connection(monkeypatch):
    monkeypatch.setattr(client, "get_connection_as_json", lambda name: "{}")
    assert client.has_connection("X") is True

    monkeypatch.setattr(client, "get_connection_as_json", lambda name: (_ for _ in ()).throw(ValueError("x")))
    assert client.has_connection("X") is False


def test_get_secret_field(monkeypatch):
    monkeypatch.setattr(client, "get_secret", lambda conn_id: {"a": {"b": 10}, "x": None})
    assert client.get_secret_field("CID", "a.b") == 10
    assert client.get_secret_field("CID", "a.c", default=5) == 5
    with pytest.raises(KeyError):
        client.get_secret_field("CID", "a.c", required=True)
    with pytest.raises(ValueError):
        client.get_secret_field("CID", "x", required=True)


def test_get_secret_field_test_mode(monkeypatch):
    monkeypatch.setattr(client, "get_secret_test", lambda conn_id, dir_path=None: {"a": 1})
    assert client.get_secret_field("CID", "a", test=True) == 1


def test_require_secret_fields(monkeypatch):
    monkeypatch.setattr(client, "get_secret", lambda conn_id: {"a": {"b": 1}, "x": "ok"})
    out = client.require_secret_fields("CID", ["a.b", "x"])
    assert out["x"] == "ok"

    monkeypatch.setattr(client, "get_secret", lambda conn_id: {"a": {"b": None}, "x": ""})
    with pytest.raises(ValueError, match="a.b"):
        client.require_secret_fields("CID", ["a.b", "x"])


def test_get_secret_with_defaults(monkeypatch):
    monkeypatch.setattr(client, "get_secret", lambda conn_id: {"b": 2})
    out = client.get_secret_with_defaults("CID", {"a": 1, "b": 0})
    assert out == {"a": 1, "b": 2}



import logging
import mimetypes
import os
import json
from typing import Dict, List, Optional

from mattermostdriver import Driver
from mnemosynecore.vault.univ_conn import un_conn


def get_mattermost_driver(bot_id: str) -> Driver:
    if bot_id.startswith('{'):
        try:
            cfg = json.loads(bot_id)
        except json.JSONDecodeError:
            raise ValueError("Неверный формат JSON для конфигурации Mattermost")

        driver = Driver({
            "url": cfg["host"],
            "token": cfg["password"],
            "scheme": cfg.get("schema", "https"),
            "port": int(cfg.get("port", 443)),
            "basepath": cfg.get("basepath", "/api/v4"),
        })
        driver.login()
        return driver

    return un_conn(bot_id, "mattermost")


def _resolve_test_bot_id(
    bot_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> str:
    from mnemosynecore.vault.client import get_secret_test

    raw = bot_id.strip()
    if raw.startswith("{"):
        try:
            json.loads(raw)
        except json.JSONDecodeError as exc:
            raise ValueError("Неверный формат JSON для конфигурации Mattermost") from exc
        return raw

    if creds_path is None:
        cfg = get_secret_test(bot_id, dir_path=dir_path)
    else:
        cfg = get_secret_test(
            bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
    return json.dumps(cfg)


def get_mattermost_driver_test(
    bot_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Driver:
    return get_mattermost_driver(
        _resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        )
    )


def send_message(
    *,
    channel_id: str,
    bot_id: str,
    text: str,
    root_id: Optional[str] = None,
    silent: bool = False,
) -> None:
    post_message(
        channel_id=channel_id,
        bot_id=bot_id,
        text=text,
        root_id=root_id,
        silent=silent,
    )


def post_message(
    *,
    channel_id: str,
    bot_id: str,
    text: str,
    root_id: Optional[str] = None,
    props: Optional[Dict[str, object]] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    driver = get_mattermost_driver(bot_id)
    options = {"channel_id": channel_id, "message": text.strip()}
    if root_id:
        options["root_id"] = root_id
    if props:
        options["props"] = props

    try:
        post = driver.posts.create_post(options=options)
        post_id = post.get("id") if isinstance(post, dict) else None
        if not silent:
            if root_id:
                logging.info(
                    "Сообщение отправлено в тред Mattermost: channel_id=%s, root_id=%s",
                    channel_id,
                    root_id,
                )
            else:
                logging.info("Сообщение отправлено в Mattermost: %s", channel_id)
        return {
            "channel_id": channel_id,
            "post_id": post_id,
            "root_id": root_id,
        }
    except Exception as exc:
        logging.exception(
            "Ошибка отправки сообщения в Mattermost (channel_id=%s)", channel_id
        )
        raise


def send_thread_message(
    *,
    channel_id: str,
    bot_id: str,
    root_id: str,
    text: str,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    if not root_id:
        raise ValueError("root_id обязателен для отправки сообщения в тред")
    return post_message(
        channel_id=channel_id,
        bot_id=bot_id,
        text=text,
        root_id=root_id,
        silent=silent,
    )


def send_thread_messages(
    *,
    channel_id: str,
    bot_id: str,
    root_id: str,
    messages: List[str],
    silent: bool = False,
    continue_on_error: bool = False,
) -> Dict[str, List[Dict[str, Optional[str]]]]:
    if not root_id:
        raise ValueError("root_id обязателен для отправки сообщений в тред")
    sent: List[Dict[str, Optional[str]]] = []
    failed: List[Dict[str, Optional[str]]] = []
    for message in messages:
        try:
            result = send_thread_message(
                channel_id=channel_id,
                bot_id=bot_id,
                root_id=root_id,
                text=message,
                silent=silent,
            )
            sent.append(result)
        except Exception as exc:
            if continue_on_error:
                failed.append(
                    {
                        "channel_id": channel_id,
                        "root_id": root_id,
                        "post_id": None,
                        "error": str(exc),
                    }
                )
                continue
            raise
    return {"sent": sent, "failed": failed}


def send_file_bytes(
    *,
    channel_id: str,
    bot_id: str,
    file_name: str,
    file_bytes: bytes,
    text: str = "",
    mime_type: Optional[str] = None,
    root_id: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    if not file_name:
        raise ValueError("file_name обязателен")
    if not isinstance(file_bytes, (bytes, bytearray)):
        raise TypeError("file_bytes должен быть bytes или bytearray")

    payload_bytes = bytes(file_bytes)
    mm_mime_type = mime_type or "application/octet-stream"

    driver = get_mattermost_driver(bot_id)
    upload = driver.files.upload_file(
        channel_id=channel_id,
        files={"files": (file_name, payload_bytes, mm_mime_type)},
    )

    file_infos = upload.get("file_infos") if isinstance(upload, dict) else None
    if not file_infos:
        raise ValueError("Mattermost не вернул file_infos после загрузки файла")

    first_file = file_infos[0]
    file_id = first_file.get("id") if isinstance(first_file, dict) else str(first_file)
    message = text.strip() if text else " "

    post = driver.posts.create_post(
        options={
            "channel_id": channel_id,
            "message": message,
            "file_ids": [file_id],
            **({"root_id": root_id} if root_id else {}),
        }
    )
    post_id = post.get("id") if isinstance(post, dict) else None

    if not silent:
        logging.info(
            "Файл отправлен в Mattermost: channel_id=%s, file_name=%s",
            channel_id,
            file_name,
        )
    return {
        "channel_id": channel_id,
        "file_name": file_name,
        "file_id": file_id,
        "post_id": post_id,
        "root_id": root_id,
    }


def send_file(
    *,
    channel_id: str,
    bot_id: str,
    file_path: str,
    text: str = "",
    file_name: Optional[str] = None,
    mime_type: Optional[str] = None,
    root_id: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(file_path)

    final_file_name = file_name or os.path.basename(file_path)
    final_mime_type = mime_type or mimetypes.guess_type(final_file_name)[0] or "application/octet-stream"

    with open(file_path, "rb") as f:
        payload = f.read()

    return send_file_bytes(
        channel_id=channel_id,
        bot_id=bot_id,
        file_name=final_file_name,
        file_bytes=payload,
        text=text,
        mime_type=final_mime_type,
        root_id=root_id,
        silent=silent,
    )


def send_files(
    *,
    channel_id: str,
    bot_id: str,
    file_paths: List[str],
    text: str = "",
    root_id: Optional[str] = None,
    silent: bool = False,
) -> List[Dict[str, Optional[str]]]:
    if not file_paths:
        return []
    results = []
    for file_path in file_paths:
        result = send_file(
            channel_id=channel_id,
            bot_id=bot_id,
            file_path=file_path,
            text=text,
            root_id=root_id,
            silent=silent,
        )
        results.append(result)
    return results


def send_dataframe_as_csv(
    *,
    channel_id: str,
    bot_id: str,
    df,
    file_name: str = "dataframe.csv",
    text: str = "",
    include_index: bool = False,
    encoding: str = "utf-8",
    root_id: Optional[str] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    import pandas as pd

    if not isinstance(df, pd.DataFrame):
        raise TypeError("df должен быть pandas.DataFrame")

    csv_bytes = df.to_csv(index=include_index).encode(encoding)
    return send_file_bytes(
        channel_id=channel_id,
        bot_id=bot_id,
        file_name=file_name,
        file_bytes=csv_bytes,
        text=text,
        mime_type="text/csv",
        root_id=root_id,
        silent=silent,
    )


def send_dataframe_preview(
    *,
    channel_id: str,
    bot_id: str,
    df,
    title: str = "",
    max_rows: int = 10,
    max_cols: int = 10,
    root_id: Optional[str] = None,
    silent: bool = False,
) -> None:
    import pandas as pd

    if not isinstance(df, pd.DataFrame):
        raise TypeError("df должен быть pandas.DataFrame")
    if max_rows < 1 or max_cols < 1:
        raise ValueError("max_rows и max_cols должны быть >= 1")

    preview = df.head(max_rows).iloc[:, :max_cols]
    preview_text = preview.to_string(index=False)
    header = f"{title.strip()}\n" if title else ""
    text = (
        f"{header}"
        f"rows={len(df)}, cols={len(df.columns)}, preview_rows={len(preview)}, preview_cols={len(preview.columns)}\n"
        f"```\n{preview_text}\n```"
    )
    send_message(
        channel_id=channel_id,
        bot_id=bot_id,
        text=text,
        root_id=root_id,
        silent=silent,
    )


def send_message_test(
    *,
    channel_id: str,
    bot_id: str,
    text: str,
    dir_path: str | None = None,
    creds_path: str | None = None,
    root_id: Optional[str] = None,
    silent: bool = False,
) -> None:
    send_message(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        text=text,
        root_id=root_id,
        silent=silent,
    )


def post_message_test(
    *,
    channel_id: str,
    bot_id: str,
    text: str,
    dir_path: str | None = None,
    creds_path: str | None = None,
    root_id: Optional[str] = None,
    props: Optional[Dict[str, object]] = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return post_message(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        text=text,
        root_id=root_id,
        props=props,
        silent=silent,
    )


def send_file_bytes_test(
    *,
    channel_id: str,
    bot_id: str,
    file_name: str,
    file_bytes: bytes,
    text: str = "",
    mime_type: Optional[str] = None,
    root_id: Optional[str] = None,
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_file_bytes(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_name=file_name,
        file_bytes=file_bytes,
        text=text,
        mime_type=mime_type,
        root_id=root_id,
        silent=silent,
    )


def send_file_test(
    *,
    channel_id: str,
    bot_id: str,
    file_path: str,
    text: str = "",
    file_name: Optional[str] = None,
    mime_type: Optional[str] = None,
    root_id: Optional[str] = None,
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_file(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_path=file_path,
        text=text,
        file_name=file_name,
        mime_type=mime_type,
        root_id=root_id,
        silent=silent,
    )


def send_files_test(
    *,
    channel_id: str,
    bot_id: str,
    file_paths: List[str],
    text: str = "",
    root_id: Optional[str] = None,
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
) -> List[Dict[str, Optional[str]]]:
    return send_files(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_paths=file_paths,
        text=text,
        root_id=root_id,
        silent=silent,
    )


def send_dataframe_as_csv_test(
    *,
    channel_id: str,
    bot_id: str,
    df,
    file_name: str = "dataframe.csv",
    text: str = "",
    include_index: bool = False,
    encoding: str = "utf-8",
    root_id: Optional[str] = None,
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_dataframe_as_csv(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        df=df,
        file_name=file_name,
        text=text,
        include_index=include_index,
        encoding=encoding,
        root_id=root_id,
        silent=silent,
    )


def send_dataframe_preview_test(
    *,
    channel_id: str,
    bot_id: str,
    df,
    title: str = "",
    max_rows: int = 10,
    max_cols: int = 10,
    root_id: Optional[str] = None,
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
) -> None:
    send_dataframe_preview(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        df=df,
        title=title,
        max_rows=max_rows,
        max_cols=max_cols,
        root_id=root_id,
        silent=silent,
    )


def send_thread_message_test(
    *,
    channel_id: str,
    bot_id: str,
    root_id: str,
    text: str,
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
) -> Dict[str, Optional[str]]:
    return send_thread_message(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        root_id=root_id,
        text=text,
        silent=silent,
    )


def send_thread_messages_test(
    *,
    channel_id: str,
    bot_id: str,
    root_id: str,
    messages: List[str],
    dir_path: str | None = None,
    creds_path: str | None = None,
    silent: bool = False,
    continue_on_error: bool = False,
) -> Dict[str, List[Dict[str, Optional[str]]]]:
    return send_thread_messages(
        channel_id=channel_id,
        bot_id=_resolve_test_bot_id(
            bot_id=bot_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        root_id=root_id,
        messages=messages,
        silent=silent,
        continue_on_error=continue_on_error,
    )



import os
import sys
import types

import pandas as pd
import pytest

from mnemosynecore import sharepoint as sp


class _Resp:
    def __init__(self, content=b"ok", status_ok=True):
        self.content = content
        self._status_ok = status_ok

    def raise_for_status(self):
        if not self._status_ok:
            raise RuntimeError("http error")


def _install_fake_ntlm(monkeypatch):
    class FakeNtlmAuth:
        def __init__(self, username, password):
            self.username = username
            self.password = password

    module = types.ModuleType("requests_ntlm")
    module.HttpNtlmAuth = FakeNtlmAuth
    monkeypatch.setitem(sys.modules, "requests_ntlm", module)
    return FakeNtlmAuth


def test_sharepoint_download_file_with_conn_id(monkeypatch):
    FakeNtlmAuth = _install_fake_ntlm(monkeypatch)
    captured = {}

    monkeypatch.setattr(
        sp,
        "_resolve_sharepoint_config",
        lambda conn, dir_path=None: {
            "host": "https://sp.local",
            "login": "svc",
            "password": "pwd",
            "domain": "O3",
        },
    )
    monkeypatch.setattr(
        sp.requests,
        "get",
        lambda url, auth=None, headers=None, timeout=None, verify=None, allow_redirects=None: (
            captured.update(
                {
                    "url": url,
                    "auth": auth,
                    "headers": headers,
                    "timeout": timeout,
                    "verify": verify,
                    "allow_redirects": allow_redirects,
                }
            )
            or _Resp(content=b"data")
        ),
    )

    out = sp.sharepoint_download_file(
        sharepoint_conn="SP_CONN",
        file_url="/sites/analytics/report.csv",
        headers={"X-Test": "1"},
        timeout=10,
        verify=False,
    )
    assert out == b"data"
    assert captured["url"] == "https://sp.local/sites/analytics/report.csv"
    assert isinstance(captured["auth"], FakeNtlmAuth)
    assert captured["auth"].username == "O3\\svc"
    assert captured["auth"].password == "pwd"
    assert captured["headers"]["X-Test"] == "1"
    assert captured["timeout"] == 10
    assert captured["verify"] is False


def test_sharepoint_download_file_absolute_url(monkeypatch):
    _install_fake_ntlm(monkeypatch)
    captured = {}

    monkeypatch.setattr(
        sp,
        "_resolve_sharepoint_config",
        lambda conn, dir_path=None: {
            "host": "https://sp.local",
            "login": "svc",
            "password": "pwd",
            "domain": None,
        },
    )
    monkeypatch.setattr(
        sp.requests,
        "get",
        lambda url, auth=None, headers=None, timeout=None, verify=None, allow_redirects=None: (
            captured.update({"url": url, "auth": auth}) or _Resp(content=b"x")
        ),
    )

    out = sp.sharepoint_download_file(
        sharepoint_conn="SP_CONN",
        file_url="https://other.local/file.txt",
    )
    assert out == b"x"
    assert captured["url"] == "https://other.local/file.txt"
    assert captured["auth"].username == "svc"


def test_sharepoint_download_file_http_error(monkeypatch):
    _install_fake_ntlm(monkeypatch)
    monkeypatch.setattr(
        sp,
        "_resolve_sharepoint_config",
        lambda conn, dir_path=None: {
            "host": "https://sp.local",
            "login": "svc",
            "password": "pwd",
        },
    )
    monkeypatch.setattr(
        sp.requests,
        "get",
        lambda *args, **kwargs: _Resp(status_ok=False),
    )
    with pytest.raises(RuntimeError, match="http error"):
        sp.sharepoint_download_file(sharepoint_conn="SP_CONN", file_url="/f")


def test_sharepoint_download_file_auth_none(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        sp,
        "_resolve_sharepoint_config",
        lambda conn, dir_path=None: {"host": "https://sp.local"},
    )
    monkeypatch.setattr(
        sp.requests,
        "get",
        lambda url, auth=None, headers=None, timeout=None, verify=None, allow_redirects=None: (
            captured.update({"auth": auth}) or _Resp(content=b"n")
        ),
    )

    out = sp.sharepoint_download_file(
        sharepoint_conn="SP_CONN",
        file_url="/f",
        auth_type="none",
    )
    assert out == b"n"
    assert captured["auth"] is None


def test_sharepoint_download_to_file(monkeypatch, tmp_path):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b"abc")
    out = sp.sharepoint_download_to_file(
        sharepoint_conn="SP_CONN",
        file_url="/x.csv",
        output_path=str(tmp_path / "sub" / "x.csv"),
    )
    assert out.endswith("x.csv")
    assert (tmp_path / "sub" / "x.csv").read_bytes() == b"abc"


def test_sharepoint_read_text(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: "привет".encode("utf-8"))
    assert sp.sharepoint_read_text(sharepoint_conn="SP", file_url="/a.txt") == "привет"


def test_sharepoint_read_dataframe_csv(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b"a,b\n1,2\n")
    df = sp.sharepoint_read_dataframe(sharepoint_conn="SP", file_url="/a.csv")
    assert list(df.columns) == ["a", "b"]
    assert df.iloc[0]["a"] == 1


def test_sharepoint_read_dataframe_tsv(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b"a\tb\n1\t2\n")
    df = sp.sharepoint_read_dataframe(sharepoint_conn="SP", file_url="/a.tsv")
    assert list(df.columns) == ["a", "b"]


def test_sharepoint_read_dataframe_json(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b'[{"a":1},{"a":2}]')
    df = sp.sharepoint_read_dataframe(sharepoint_conn="SP", file_url="/a.json")
    assert list(df["a"]) == [1, 2]


def test_sharepoint_read_dataframe_excel(monkeypatch):
    captured = {}

    def fake_read_excel(buffer, **kwargs):
        captured["is_bytesio"] = hasattr(buffer, "read")
        return pd.DataFrame({"a": [1], "b": [2]})

    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b"excel-bytes")
    monkeypatch.setattr(sp.pd, "read_excel", fake_read_excel)
    df = sp.sharepoint_read_dataframe(sharepoint_conn="SP", file_url="/x.xlsx")
    assert list(df.columns) == ["a", "b"]
    assert captured["is_bytesio"] is True


def test_sharepoint_read_dataframe_unknown_auto(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b"x")
    with pytest.raises(ValueError, match="Не удалось определить формат"):
        sp.sharepoint_read_dataframe(sharepoint_conn="SP", file_url="/x.unknown")


def test_sharepoint_read_dataframe_unknown_explicit(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_download_file", lambda **kwargs: b"x")
    with pytest.raises(ValueError, match="file_format должен быть"):
        sp.sharepoint_read_dataframe(sharepoint_conn="SP", file_url="/x.any", file_format="parquet")


def test_sharepoint_read_csv_wrapper(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        sp,
        "sharepoint_read_dataframe",
        lambda **kwargs: captured.update(kwargs) or pd.DataFrame({"a": [1]}),
    )
    df = sp.sharepoint_read_csv(sharepoint_conn="SP", file_url="/a.csv")
    assert list(df["a"]) == [1]
    assert captured["file_format"] == "csv"


def test_sharepoint_read_excel_wrapper(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        sp,
        "sharepoint_read_dataframe",
        lambda **kwargs: captured.update(kwargs) or pd.DataFrame({"a": [1]}),
    )
    sp.sharepoint_read_excel(sharepoint_conn="SP", file_url="/a.xlsx")
    assert captured["file_format"] == "excel"


def test_sharepoint_read_json_wrapper(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        sp,
        "sharepoint_read_dataframe",
        lambda **kwargs: captured.update(kwargs) or pd.DataFrame({"a": [1]}),
    )
    sp.sharepoint_read_json(sharepoint_conn="SP", file_url="/a.json")
    assert captured["file_format"] == "json"


def test_sharepoint_read_sql_wrapper(monkeypatch):
    monkeypatch.setattr(sp, "sharepoint_read_text", lambda **kwargs: "select 1;")
    assert sp.sharepoint_read_sql(sharepoint_conn="SP", file_url="/q.sql") == "select 1;"


def test_sharepoint_download_many(monkeypatch, tmp_path):
    calls = []
    monkeypatch.setattr(
        sp,
        "sharepoint_download_to_file",
        lambda **kwargs: calls.append(kwargs) or kwargs["output_path"],
    )
    out = sp.sharepoint_download_many(
        sharepoint_conn="SP",
        files=[
            {"file_url": "/a.csv"},
            {"url": "/b.csv", "file_name": "b2.csv"},
        ],
        output_dir=str(tmp_path),
    )
    assert len(out["downloaded"]) == 2
    assert out["failed"] == []
    assert calls[1]["output_path"].endswith("b2.csv")


def test_sharepoint_download_many_continue_on_error(monkeypatch, tmp_path):
    def fake_download(**kwargs):
        if kwargs["file_url"] == "/bad.csv":
            raise RuntimeError("bad")
        return kwargs["output_path"]

    monkeypatch.setattr(sp, "sharepoint_download_to_file", fake_download)
    out = sp.sharepoint_download_many(
        sharepoint_conn="SP",
        files=[{"file_url": "/ok.csv"}, {"file_url": "/bad.csv"}],
        output_dir=str(tmp_path),
        continue_on_error=True,
    )
    assert len(out["downloaded"]) == 1
    assert len(out["failed"]) == 1


def test_sharepoint_download_many_requires_url(tmp_path):
    with pytest.raises(ValueError):
        sp.sharepoint_download_many(
            sharepoint_conn="SP",
            files=[{"file_name": "x.csv"}],
            output_dir=str(tmp_path),
        )


def test_sharepoint_read_csv_test_supports_creds_path(monkeypatch, tmp_path):
    creds_file = tmp_path / "sp_creds.json"
    creds_file.write_text(
        '{"host":"https://sp.local","login":"svc","password":"pwd"}',
        encoding="utf-8",
    )
    captured = {}

    monkeypatch.setattr(
        sp,
        "sharepoint_read_csv",
        lambda **kwargs: captured.update(kwargs) or pd.DataFrame({"a": [1]}),
    )

    df = sp.sharepoint_read_csv_test(
        sharepoint_conn="SP_CONN",
        file_url="/a.csv",
        creds_path=str(creds_file),
    )
    assert list(df["a"]) == [1]
    assert captured["sharepoint_conn"]["host"] == "https://sp.local"



from mnemosynecore import secrets


def test_resolve_secret_prefers_test_file(monkeypatch):
    monkeypatch.setattr(secrets, "get_secret_test", lambda conn_id, dir_path=None: {"from": "test"})
    monkeypatch.setattr(secrets, "get_secret", lambda conn_id: {"from": "prod"})
    assert secrets.resolve_secret("CID") == {"from": "test"}


def test_resolve_secret_fallback_to_prod(monkeypatch):
    def fail(*_args, **_kwargs):
        raise FileNotFoundError()

    monkeypatch.setattr(secrets, "get_secret_test", fail)
    monkeypatch.setattr(secrets, "get_secret", lambda conn_id: {"from": "prod"})
    assert secrets.resolve_secret("CID") == {"from": "prod"}


def test_resolve_secret_test_supports_creds_path(monkeypatch):
    captured = {}

    def fake_get_secret_test(conn_id, dir_path=None, creds_path=None):
        captured["creds_path"] = creds_path
        return {"from": "test"}

    monkeypatch.setattr(secrets, "get_secret_test", fake_get_secret_test)
    assert secrets.resolve_secret_test("CID", creds_path="/tmp/cid.json") == {"from": "test"}
    assert captured["creds_path"] == "/tmp/cid.json"



import io
import json
import os
from typing import Any, Dict, Optional, Union
from urllib.parse import urljoin

import pandas as pd
import requests
from mnemosynecore.vault.client import get_secret_test

SharePointConn = Union[str, Dict[str, Any]]


def _parse_extra(extra_raw: Any) -> Dict[str, Any]:
    if isinstance(extra_raw, dict):
        return extra_raw
    if isinstance(extra_raw, str):
        try:
            parsed = json.loads(extra_raw) if extra_raw else {}
            if isinstance(parsed, dict):
                return parsed
        except json.JSONDecodeError:
            return {}
    return {}


def _resolve_sharepoint_config(
    sharepoint_conn: SharePointConn,
    dir_path: Optional[str] = None,
) -> Dict[str, Any]:
    if isinstance(sharepoint_conn, dict):
        cfg = dict(sharepoint_conn)
    elif isinstance(sharepoint_conn, str):
        raw = sharepoint_conn.strip()
        if raw.startswith("{"):
            try:
                cfg = json.loads(raw)
            except json.JSONDecodeError as exc:
                raise ValueError("Неверный формат JSON для конфигурации SharePoint") from exc
        else:
            from mnemosynecore.secrets import resolve_secret

            cfg = resolve_secret(raw, dir_path)
    else:
        raise TypeError("sharepoint_conn должен быть dict или str")

    extra = _parse_extra(cfg.get("extra"))
    out = dict(cfg)
    out["extra"] = extra
    out["host"] = out.get("host") or out.get("url") or out.get("sharepoint_url") or extra.get("host")
    out["login"] = out.get("login") or out.get("username") or out.get("user") or extra.get("login")
    out["password"] = out.get("password") or extra.get("password")
    out["domain"] = out.get("domain") or out.get("schema") or extra.get("domain")
    return out


def _resolve_sharepoint_config_test(
    sharepoint_conn: SharePointConn,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict[str, Any]:
    if isinstance(sharepoint_conn, dict):
        return _resolve_sharepoint_config(sharepoint_conn, dir_path=None)
    if isinstance(sharepoint_conn, str):
        raw = sharepoint_conn.strip()
        if raw.startswith("{"):
            return _resolve_sharepoint_config(raw, dir_path=None)
        return _resolve_sharepoint_config(
            get_secret_test(raw, dir_path=dir_path, creds_path=creds_path),
            dir_path=None,
        )
    raise TypeError("sharepoint_conn должен быть dict или str")


def _build_request_url(host: str, file_url: str) -> str:
    file_url = file_url.strip()
    if file_url.startswith("http://") or file_url.startswith("https://"):
        return file_url
    return urljoin(host.rstrip("/") + "/", file_url.lstrip("/"))


def _build_ntlm_username(login: str, domain: Optional[str]) -> str:
    if not domain:
        return login
    if "\\" in login or "@" in login:
        return login
    return f"{domain}\\{login}"


def sharepoint_download_file(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> bytes:
    cfg = _resolve_sharepoint_config(sharepoint_conn, dir_path)

    host = cfg.get("host")
    login = cfg.get("login")
    password = cfg.get("password")
    domain = cfg.get("domain")

    if not host:
        raise ValueError("Для SharePoint нужен host/sharepoint_url")
    if auth_type != "none":
        if not login:
            raise ValueError("Для SharePoint нужен login/username")
        if password is None:
            raise ValueError("Для SharePoint нужен password")

    url = _build_request_url(host, file_url)
    request_headers = dict(headers or {})

    auth = None
    if auth_type.lower() == "ntlm":
        try:
            from requests_ntlm import HttpNtlmAuth
        except ImportError as exc:
            raise ImportError(
                "Для NTLM-доступа к SharePoint установите requests-ntlm "
                "(или mnemosynecore[sharepoint])"
            ) from exc
        auth = HttpNtlmAuth(_build_ntlm_username(login, domain), password)
    elif auth_type.lower() == "basic":
        auth = (login, password)
    elif auth_type.lower() == "none":
        auth = None
    else:
        raise ValueError("auth_type должен быть одним из: ntlm, basic, none")

    response = requests.get(
        url,
        auth=auth,
        headers=request_headers,
        timeout=timeout,
        verify=verify,
        allow_redirects=allow_redirects,
    )
    response.raise_for_status()
    return response.content


def sharepoint_download_to_file(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    output_path: str,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> str:
    content = sharepoint_download_file(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )

    out_dir = os.path.dirname(output_path)
    if out_dir:
        os.makedirs(out_dir, exist_ok=True)

    with open(output_path, "wb") as f:
        f.write(content)
    return output_path


def sharepoint_read_text(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    errors: str = "strict",
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> str:
    content = sharepoint_download_file(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )
    return content.decode(encoding, errors=errors)


def sharepoint_read_dataframe(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    file_format: str = "auto",
    encoding: str = "utf-8",
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    content = sharepoint_download_file(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )

    options = dict(read_kwargs or {})
    fmt = file_format.lower().strip()
    if fmt == "auto":
        path_lower = file_url.lower()
        if path_lower.endswith(".csv"):
            fmt = "csv"
        elif path_lower.endswith(".tsv"):
            fmt = "tsv"
        elif path_lower.endswith(".xlsx") or path_lower.endswith(".xls"):
            fmt = "excel"
        elif path_lower.endswith(".json"):
            fmt = "json"
        else:
            raise ValueError(
                "Не удалось определить формат файла автоматически. "
                "Передайте file_format явно (csv/tsv/excel/json)."
            )

    if fmt == "csv":
        return pd.read_csv(io.StringIO(content.decode(encoding)), **options)
    if fmt == "tsv":
        if "sep" not in options:
            options["sep"] = "\t"
        return pd.read_csv(io.StringIO(content.decode(encoding)), **options)
    if fmt == "excel":
        return pd.read_excel(io.BytesIO(content), **options)
    if fmt == "json":
        return pd.read_json(io.StringIO(content.decode(encoding)), **options)

    raise ValueError("file_format должен быть одним из: auto, csv, tsv, excel, json")


def sharepoint_read_csv(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_dataframe(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        file_format="csv",
        encoding=encoding,
        read_kwargs=read_kwargs,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_excel(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_dataframe(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        file_format="excel",
        read_kwargs=read_kwargs,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_json(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_dataframe(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        file_format="json",
        encoding=encoding,
        read_kwargs=read_kwargs,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_sql(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> str:
    return sharepoint_read_text(
        sharepoint_conn=sharepoint_conn,
        file_url=file_url,
        encoding=encoding,
        dir_path=dir_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_download_many(
    *,
    sharepoint_conn: SharePointConn,
    files: list,
    output_dir: str,
    dir_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
    continue_on_error: bool = False,
) -> Dict[str, Any]:
    os.makedirs(output_dir, exist_ok=True)
    downloaded = []
    failed = []

    for i, item in enumerate(files, start=1):
        file_url = item.get("file_url") or item.get("url")
        if not file_url:
            err = "file_url/url обязателен"
            if continue_on_error:
                failed.append({"item": item, "error": err})
                continue
            raise ValueError(err)

        file_name = item.get("file_name")
        if not file_name:
            file_name = os.path.basename(file_url.strip("/")) or f"file_{i}"
        output_path = os.path.join(output_dir, file_name)

        try:
            saved_path = sharepoint_download_to_file(
                sharepoint_conn=sharepoint_conn,
                file_url=file_url,
                output_path=output_path,
                dir_path=dir_path,
                timeout=timeout,
                verify=verify,
                headers=headers,
                auth_type=auth_type,
                allow_redirects=allow_redirects,
            )
            downloaded.append(saved_path)
        except Exception as exc:
            if continue_on_error:
                failed.append({"item": item, "error": str(exc)})
                continue
            raise

    return {"downloaded": downloaded, "failed": failed}


def sharepoint_download_file_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> bytes:
    return sharepoint_download_file(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_download_to_file_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    output_path: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> str:
    return sharepoint_download_to_file(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        output_path=output_path,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_text_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    errors: str = "strict",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> str:
    return sharepoint_read_text(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        encoding=encoding,
        errors=errors,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_dataframe_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    file_format: str = "auto",
    encoding: str = "utf-8",
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_dataframe(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        file_format=file_format,
        encoding=encoding,
        read_kwargs=read_kwargs,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_csv_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_csv(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        encoding=encoding,
        read_kwargs=read_kwargs,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_excel_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_excel(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        read_kwargs=read_kwargs,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_json_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    read_kwargs: Optional[Dict[str, Any]] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> pd.DataFrame:
    return sharepoint_read_json(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        encoding=encoding,
        read_kwargs=read_kwargs,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_read_sql_test(
    *,
    sharepoint_conn: SharePointConn,
    file_url: str,
    encoding: str = "utf-8",
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
) -> str:
    return sharepoint_read_sql(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        file_url=file_url,
        encoding=encoding,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
    )


def sharepoint_download_many_test(
    *,
    sharepoint_conn: SharePointConn,
    files: list,
    output_dir: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
    timeout: int = 60,
    verify: bool = True,
    headers: Optional[Dict[str, str]] = None,
    auth_type: str = "ntlm",
    allow_redirects: bool = True,
    continue_on_error: bool = False,
) -> Dict[str, Any]:
    return sharepoint_download_many(
        sharepoint_conn=_resolve_sharepoint_config_test(
            sharepoint_conn=sharepoint_conn,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        files=files,
        output_dir=output_dir,
        timeout=timeout,
        verify=verify,
        headers=headers,
        auth_type=auth_type,
        allow_redirects=allow_redirects,
        continue_on_error=continue_on_error,
    )



import warnings

import mnemosynecore as core
from mnemosynecore.warnings import old_function


def test_old_function_emits_deprecation_warning():
    with warnings.catch_warnings(record=True) as recorded:
        warnings.simplefilter("always")
        old_function()
    assert recorded
    assert issubclass(recorded[0].category, DeprecationWarning)


def test_init_exports_are_accessible():
    for name in core.__all__:
        assert hasattr(core, name), f"Missing export: {name}"




from mnemosynecore.data_bases.vertica import (
    vertica_conn,
    vertica_conn_test,
    vertica_sql,
    vertica_sql_test,
    vertica_select,
    vertica_select_test,
    vertica_select_scalar,
    vertica_select_scalar_test,
    load_sql_tasks_from_dir,
    read_sql_file,
    split_sql_statements,
    vertica_sql_file,
    vertica_sql_file_test,
    vertica_sql_dir,
    vertica_sql_dir_test,
    vertica_insert_dataframe,
    vertica_insert_dataframe_test,
    vertica_table_exists,
    vertica_table_exists_test,
    vertica_dedupe,
    vertica_dedupe_test,
    vertica_upsert,
    vertica_upsert_test,
    get_vertica_engine_test,
)
from .mattermost import (
    get_mattermost_driver,
    get_mattermost_driver_test,
    post_message,
    post_message_test,
    send_message,
    send_message_test,
    send_thread_message,
    send_thread_message_test,
    send_thread_messages,
    send_thread_messages_test,
    send_file,
    send_file_test,
    send_file_bytes,
    send_file_bytes_test,
    send_files,
    send_files_test,
    send_dataframe_as_csv,
    send_dataframe_as_csv_test,
    send_dataframe_preview,
    send_dataframe_preview_test,
)
from .superset import (
    superset_request,
    superset_request_test,
    superset_dashboard_url,
    superset_dashboard_url_test,
    superset_chart_url,
    superset_chart_url_test,
    superset_dashboard_link,
    superset_dashboard_link_test,
    superset_chart_link,
    superset_chart_link_test,
    get_superset_client,
    get_superset_client_test,
    superset_dashboard_thumbnail,
    superset_dashboard_thumbnail_test,
    superset_chart_thumbnail,
    superset_chart_thumbnail_test,
    superset_dashboard_screenshot_to_file,
    superset_dashboard_screenshot_to_file_test,
    superset_chart_screenshot_to_file,
    superset_chart_screenshot_to_file_test,
    superset_screenshot_dashboard,
    superset_screenshot_dashboard_test,
    superset_screenshot_charts,
    superset_screenshot_charts_test,
    send_superset_dashboard_screenshot,
    send_superset_dashboard_screenshot_test,
    send_superset_dashboard_permalink_screenshot,
    send_superset_dashboard_permalink_screenshot_test,
    send_superset_chart_screenshot,
    send_superset_chart_screenshot_test,
    send_superset_dashboard_link,
    send_superset_dashboard_link_test,
    send_superset_chart_link,
    send_superset_chart_link_test,
    send_superset_dashboard_links_to_thread,
    send_superset_dashboard_links_to_thread_test,
    send_superset_chart_links_to_thread,
    send_superset_chart_links_to_thread_test,
    send_superset_dashboards_to_channels,
    send_superset_dashboards_to_channels_test,
    send_superset_charts_to_channels,
    send_superset_charts_to_channels_test,
)
from .sharepoint import (
    sharepoint_download_file,
    sharepoint_download_file_test,
    sharepoint_download_to_file,
    sharepoint_download_to_file_test,
    sharepoint_read_text,
    sharepoint_read_text_test,
    sharepoint_read_dataframe,
    sharepoint_read_dataframe_test,
    sharepoint_read_csv,
    sharepoint_read_csv_test,
    sharepoint_read_excel,
    sharepoint_read_excel_test,
    sharepoint_read_json,
    sharepoint_read_json_test,
    sharepoint_read_sql,
    sharepoint_read_sql_test,
    sharepoint_download_many,
    sharepoint_download_many_test,
)
from .retries import retry_call, retry
from mnemosynecore.vault.client import get_secret, get_secret_test, get_connection_as_json, get_connection_as_json_test
from mnemosynecore.vault.client import (
    has_connection,
    get_secret_field,
    require_secret_fields,
    get_secret_with_defaults,
)
from mnemosynecore.vault.univ_conn import un_conn, un_conn_test
from .secrets import resolve_secret, resolve_secret_test
from .warnings import old_function


__all__ = [
    "vertica_conn",
    "vertica_conn_test",
    "get_vertica_engine_test",
    "un_conn",
    "un_conn_test",
    "load_sql_tasks_from_dir",
    "read_sql_file",
    "split_sql_statements",
    "vertica_sql_file",
    "vertica_sql_file_test",
    "vertica_sql_dir",
    "vertica_sql_dir_test",
    "vertica_insert_dataframe",
    "vertica_insert_dataframe_test",
    "vertica_table_exists",
    "vertica_table_exists_test",
    "vertica_dedupe",
    "vertica_dedupe_test",
    "vertica_upsert",
    "vertica_upsert_test",
    "vertica_sql",
    "vertica_sql_test",
    "vertica_select",
    "vertica_select_test",
    "vertica_select_scalar",
    "vertica_select_scalar_test",
    "get_mattermost_driver",
    "get_mattermost_driver_test",
    "post_message",
    "post_message_test",
    "send_message",
    "send_message_test",
    "send_thread_message",
    "send_thread_message_test",
    "send_thread_messages",
    "send_thread_messages_test",
    "send_file",
    "send_file_test",
    "send_file_bytes",
    "send_file_bytes_test",
    "send_files",
    "send_files_test",
    "send_dataframe_as_csv",
    "send_dataframe_as_csv_test",
    "send_dataframe_preview",
    "send_dataframe_preview_test",
    "superset_request",
    "superset_request_test",
    "superset_dashboard_url",
    "superset_dashboard_url_test",
    "superset_chart_url",
    "superset_chart_url_test",
    "superset_dashboard_link",
    "superset_dashboard_link_test",
    "superset_chart_link",
    "superset_chart_link_test",
    "get_superset_client",
    "get_superset_client_test",
    "superset_dashboard_thumbnail",
    "superset_dashboard_thumbnail_test",
    "superset_chart_thumbnail",
    "superset_chart_thumbnail_test",
    "superset_dashboard_screenshot_to_file",
    "superset_dashboard_screenshot_to_file_test",
    "superset_chart_screenshot_to_file",
    "superset_chart_screenshot_to_file_test",
    "get_secret",
    "get_secret_test",
    "get_connection_as_json",
    "get_connection_as_json_test",
    "has_connection",
    "get_secret_field",
    "require_secret_fields",
    "get_secret_with_defaults",
    "resolve_secret_test",
    "superset_screenshot_dashboard",
    "superset_screenshot_dashboard_test",
    "superset_screenshot_charts",
    "superset_screenshot_charts_test",
    "send_superset_dashboard_screenshot",
    "send_superset_dashboard_screenshot_test",
    "send_superset_dashboard_permalink_screenshot",
    "send_superset_dashboard_permalink_screenshot_test",
    "send_superset_chart_screenshot",
    "send_superset_chart_screenshot_test",
    "send_superset_dashboard_link",
    "send_superset_dashboard_link_test",
    "send_superset_chart_link",
    "send_superset_chart_link_test",
    "send_superset_dashboard_links_to_thread",
    "send_superset_dashboard_links_to_thread_test",
    "send_superset_chart_links_to_thread",
    "send_superset_chart_links_to_thread_test",
    "send_superset_dashboards_to_channels",
    "send_superset_dashboards_to_channels_test",
    "send_superset_charts_to_channels",
    "send_superset_charts_to_channels_test",
    "sharepoint_download_file",
    "sharepoint_download_file_test",
    "sharepoint_download_to_file",
    "sharepoint_download_to_file_test",
    "sharepoint_read_text",
    "sharepoint_read_text_test",
    "sharepoint_read_dataframe",
    "sharepoint_read_dataframe_test",
    "sharepoint_read_csv",
    "sharepoint_read_csv_test",
    "sharepoint_read_excel",
    "sharepoint_read_excel_test",
    "sharepoint_read_json",
    "sharepoint_read_json_test",
    "sharepoint_read_sql",
    "sharepoint_read_sql_test",
    "sharepoint_download_many",
    "sharepoint_download_many_test",
    "retry_call",
    "retry",
    "resolve_secret",
    "old_function",
]



import pytest

from mnemosynecore.retries import retry, retry_call


def test_retry_call_succeeds_after_retries():
    state = {"n": 0}
    sleeps = []
    retries = []

    def func():
        state["n"] += 1
        if state["n"] < 3:
            raise ValueError("fail")
        return "ok"

    out = retry_call(
        func,
        attempts=4,
        delay_sec=0.1,
        backoff=2.0,
        sleep_func=lambda sec: sleeps.append(sec),
        on_retry=lambda attempt, exc, sec: retries.append((attempt, str(exc), sec)),
    )

    assert out == "ok"
    assert state["n"] == 3
    assert sleeps == [0.1, 0.2]
    assert retries[0][0] == 1 and retries[1][0] == 2


def test_retry_call_raises_last_exception():
    state = {"n": 0}

    def func():
        state["n"] += 1
        raise RuntimeError("boom")

    with pytest.raises(RuntimeError, match="boom"):
        retry_call(func, attempts=3, sleep_func=lambda sec: None)
    assert state["n"] == 3


def test_retry_call_validates_arguments():
    with pytest.raises(ValueError):
        retry_call(lambda: 1, attempts=0)
    with pytest.raises(ValueError):
        retry_call(lambda: 1, delay_sec=-1)
    with pytest.raises(ValueError):
        retry_call(lambda: 1, backoff=0.5)
    with pytest.raises(ValueError):
        retry_call(lambda: 1, exceptions=[])


def test_retry_decorator():
    state = {"n": 0}

    @retry(attempts=2, delay_sec=0, sleep_func=lambda sec: None)
    def f():
        state["n"] += 1
        if state["n"] == 1:
            raise ValueError("once")
        return "done"

    assert f() == "done"
    assert state["n"] == 2




import random
import time
from typing import Any, Callable, Optional, Sequence, Tuple, Type


def retry_call(
    func: Callable[..., Any],
    *args,
    attempts: int = 3,
    delay_sec: float = 1.0,
    backoff: float = 2.0,
    max_delay_sec: Optional[float] = None,
    jitter_sec: float = 0.0,
    exceptions: Sequence[Type[BaseException]] = (Exception,),
    on_retry: Optional[Callable[[int, BaseException, float], None]] = None,
    sleep_func: Callable[[float], None] = time.sleep,
    **kwargs,
) -> Any:
    if attempts < 1:
        raise ValueError("attempts должно быть >= 1")
    if delay_sec < 0:
        raise ValueError("delay_sec должно быть >= 0")
    if backoff < 1:
        raise ValueError("backoff должно быть >= 1")
    if jitter_sec < 0:
        raise ValueError("jitter_sec должно быть >= 0")
    if max_delay_sec is not None and max_delay_sec < 0:
        raise ValueError("max_delay_sec должно быть >= 0")

    retry_exceptions: Tuple[Type[BaseException], ...] = tuple(exceptions)
    if not retry_exceptions:
        raise ValueError("exceptions не должен быть пустым")

    current_delay = float(delay_sec)
    last_exc: Optional[BaseException] = None

    for attempt in range(1, attempts + 1):
        try:
            return func(*args, **kwargs)
        except retry_exceptions as exc:
            last_exc = exc
            if attempt >= attempts:
                break

            sleep_for = current_delay
            if jitter_sec > 0:
                sleep_for += random.uniform(0, jitter_sec)
            if max_delay_sec is not None:
                sleep_for = min(sleep_for, max_delay_sec)

            if on_retry:
                on_retry(attempt, exc, sleep_for)
            if sleep_for > 0:
                sleep_func(sleep_for)

            next_delay = current_delay * backoff
            if max_delay_sec is not None:
                next_delay = min(next_delay, max_delay_sec)
            current_delay = next_delay

    if last_exc is not None:
        raise last_exc
    raise RuntimeError("retry_call: неизвестная ошибка выполнения")


def retry(
    *,
    attempts: int = 3,
    delay_sec: float = 1.0,
    backoff: float = 2.0,
    max_delay_sec: Optional[float] = None,
    jitter_sec: float = 0.0,
    exceptions: Sequence[Type[BaseException]] = (Exception,),
    on_retry: Optional[Callable[[int, BaseException, float], None]] = None,
    sleep_func: Callable[[float], None] = time.sleep,
):
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        def wrapped(*args, **kwargs):
            return retry_call(
                func,
                *args,
                attempts=attempts,
                delay_sec=delay_sec,
                backoff=backoff,
                max_delay_sec=max_delay_sec,
                jitter_sec=jitter_sec,
                exceptions=exceptions,
                on_retry=on_retry,
                sleep_func=sleep_func,
                **kwargs,
            )

        wrapped.__name__ = getattr(func, "__name__", "wrapped")
        wrapped.__doc__ = getattr(func, "__doc__")
        return wrapped

    return decorator



import sys
import types

import pandas as pd
import pytest

from mnemosynecore.data_bases import vertica


class DummyCursor:
    def __init__(self, fail=False):
        self.fail = fail
        self.calls = []
        self.description = [("id",), ("name",)]
        self._rows = [(1, "alpha"), (2, "beta")]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def execute(self, sql, params=None):
        if self.fail:
            raise RuntimeError("sql-fail")
        self.calls.append(("execute", sql, params))

    def executemany(self, sql, params=None):
        if self.fail:
            raise RuntimeError("sql-fail")
        self.calls.append(("executemany", sql, params))

    def fetchall(self):
        return self._rows

    def fetchone(self):
        return self._rows[0] if self._rows else None


class DummyConn:
    def __init__(self, fail=False):
        self.cursor_obj = DummyCursor(fail=fail)
        self.commit_calls = 0
        self.rollback_calls = 0
        self.close_calls = 0

    def cursor(self):
        return self.cursor_obj

    def commit(self):
        self.commit_calls += 1

    def rollback(self):
        self.rollback_calls += 1

    def close(self):
        self.close_calls += 1


def test_read_sql_file_returns_content(tmp_path):
    sql_file = tmp_path / "q.sql"
    sql_file.write_text("select 1;", encoding="utf-8")
    assert vertica.read_sql_file(str(sql_file)) == "select 1;"


def test_read_sql_file_missing_returns_none(tmp_path):
    assert vertica.read_sql_file(str(tmp_path / "none.sql")) is None


def test_load_sql_tasks_from_dir_raises_without_airflow(monkeypatch):
    monkeypatch.setattr(vertica, "HAS_AIRFLOW", False)
    with pytest.raises(ImportError):
        vertica.load_sql_tasks_from_dir("/tmp", "CID")


def test_load_sql_tasks_from_dir_success(monkeypatch, tmp_path):
    class FakeVerticaOperator:
        def __init__(self, **kwargs):
            self.kwargs = kwargs

    module = types.ModuleType("airflow.providers.vertica.operators.vertica")
    module.VerticaOperator = FakeVerticaOperator
    monkeypatch.setitem(sys.modules, "airflow.providers.vertica.operators.vertica", module)
    monkeypatch.setattr(vertica, "HAS_AIRFLOW", True)
    monkeypatch.setattr(vertica, "get_current_context", lambda: {"dag": "DAG1"}, raising=False)

    (tmp_path / "a.sql").write_text("select 1;select 2;", encoding="utf-8")
    (tmp_path / "b.txt").write_text("skip", encoding="utf-8")

    tasks = vertica.load_sql_tasks_from_dir(str(tmp_path), "VERT")
    assert "task_a_vertica" in tasks
    assert tasks["task_a_vertica"].kwargs["vertica_conn_id"] == "VERT"


def test_vertica_dedupe_requires_conn_or_id():
    with pytest.raises(ValueError):
        vertica.vertica_dedupe("t", ["k1"])


def test_vertica_dedupe_builds_sql_for_all_keys(monkeypatch):
    captured = {}

    def fake_vertica_sql(**kwargs):
        captured.update(kwargs)

    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_sql", fake_vertica_sql)
    vertica.vertica_dedupe(table_name="tbl", unique_keys=["k1", "k2"], conn=conn, date_col="dt")
    sql = captured["sql"]
    assert "t.k1 = x.k1 AND t.k2 = x.k2" in sql
    assert "ORDER BY dt DESC" in sql


def test_vertica_get_connection_as_json_delegates(monkeypatch):
    monkeypatch.setattr(vertica, "_get_connection_as_json", lambda name: "ok-json")
    assert vertica.get_connection_as_json("CID") == "ok-json"


def test_vertica_upsert_empty_df_returns_early():
    conn = DummyConn()
    vertica.vertica_upsert(pd.DataFrame(), "tbl", "id", conn=conn)
    assert conn.commit_calls == 0


def test_vertica_upsert_with_conn(monkeypatch):
    calls = []

    def fake_vertica_sql(**kwargs):
        calls.append(kwargs)

    conn = DummyConn()
    df = pd.DataFrame([{"id": 1, "dt": "2026-01-01", "val": 10}])
    monkeypatch.setattr(vertica, "vertica_sql", fake_vertica_sql)
    vertica.vertica_upsert(df=df, table_name="tbl", unique_keys="id", conn=conn, date_col="dt", days_back=7, commit=True)

    assert len(calls) == 4
    assert conn.commit_calls == 1
    assert conn.close_calls == 0


def test_vertica_upsert_with_conn_id_closes_conn(monkeypatch):
    calls = []
    conn = DummyConn()
    df = pd.DataFrame([{"id": 1, "val": 10}])
    monkeypatch.setattr(vertica, "vertica_sql", lambda **kwargs: calls.append(kwargs))
    monkeypatch.setattr(vertica, "vertica_conn", lambda conn_id: conn)
    vertica.vertica_upsert(df=df, table_name="tbl", unique_keys="id", conn_id="CID", commit=False)

    assert len(calls) == 3
    assert conn.commit_calls == 0
    assert conn.close_calls == 1


def test_vertica_conn_uses_secret_and_extra(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        vertica,
        "get_secret",
        lambda conn_id: {
            "host": "vh",
            "port": 5433,
            "login": "u",
            "password": "p",
            "schema": "db",
            "extra": '{"session_label":"x"}',
        },
    )
    monkeypatch.setattr(vertica.vertica_python, "connect", lambda **kwargs: captured.setdefault("kwargs", kwargs) or "CONN")
    out = vertica.vertica_conn("CID")
    assert out == captured["kwargs"]
    assert captured["kwargs"]["database"] == "db"
    assert captured["kwargs"]["session_label"] == "x"


def test_vertica_conn_supports_user_field(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        vertica,
        "get_secret",
        lambda conn_id: {
            "host": "vh",
            "port": 5433,
            "user": "from_user",
            "password": "p",
            "schema": "db",
        },
    )
    monkeypatch.setattr(
        vertica.vertica_python,
        "connect",
        lambda **kwargs: captured.setdefault("kwargs", kwargs) or "CONN",
    )
    vertica.vertica_conn("CID")
    assert captured["kwargs"]["user"] == "from_user"


def test_get_vertica_engine(monkeypatch):
    monkeypatch.setattr(vertica, "un_conn", lambda conn_id, conn_type: ("E", conn_id, conn_type))
    assert vertica.get_vertica_engine("CID") == ("E", "CID", "vertica")


def test_vertica_sql_requires_conn_or_id():
    with pytest.raises(ValueError):
        vertica.vertica_sql(sql="select 1")


def test_vertica_sql_execute_and_commit():
    conn = DummyConn()
    vertica.vertica_sql(conn=conn, sql="select 1", params={"a": 1}, commit=True)
    assert conn.cursor_obj.calls[0][0] == "execute"
    assert conn.commit_calls == 1


def test_vertica_sql_executemany():
    conn = DummyConn()
    rows = [(1,), (2,)]
    vertica.vertica_sql(conn=conn, sql="insert", params=rows, many=True)
    assert conn.cursor_obj.calls[0][0] == "executemany"


def test_vertica_sql_rollbacks_and_reraises():
    conn = DummyConn(fail=True)
    with pytest.raises(RuntimeError, match="sql-fail"):
        vertica.vertica_sql(conn=conn, sql="broken")
    assert conn.rollback_calls == 1


def test_vertica_sql_conn_id_closes(monkeypatch):
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_conn", lambda conn_id: conn)
    vertica.vertica_sql(conn_id="CID", sql="select 1")
    assert conn.close_calls == 1


def test_vertica_select_requires_conn_or_id():
    with pytest.raises(ValueError):
        vertica.vertica_select(sql="select 1")


def test_vertica_select_returns_dataframe():
    conn = DummyConn()
    df = vertica.vertica_select(conn=conn, sql="select * from t")
    assert list(df.columns) == ["id", "name"]
    assert df.shape == (2, 2)


def test_vertica_select_conn_id_closes(monkeypatch):
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_conn", lambda conn_id: conn)
    vertica.vertica_select(conn_id="CID", sql="select 1")
    assert conn.close_calls == 1


def test_split_sql_statements():
    assert vertica.split_sql_statements("select 1; select 2;") == ["select 1", "select 2"]
    assert vertica.split_sql_statements("") == []


def test_vertica_sql_file_not_found(tmp_path):
    with pytest.raises(FileNotFoundError):
        vertica.vertica_sql_file(file_path=str(tmp_path / "x.sql"), conn=DummyConn())


def test_vertica_sql_file_executes_statements(monkeypatch, tmp_path):
    sql_file = tmp_path / "q.sql"
    sql_file.write_text("select 1;select 2;", encoding="utf-8")
    calls = []
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_sql", lambda **kwargs: calls.append(kwargs))
    executed = vertica.vertica_sql_file(file_path=str(sql_file), conn=conn)
    assert executed == 2
    assert len(calls) == 2
    assert conn.commit_calls == 1


def test_vertica_sql_file_uses_conn_id(monkeypatch, tmp_path):
    sql_file = tmp_path / "q.sql"
    sql_file.write_text("select 1;", encoding="utf-8")
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_conn", lambda conn_id: conn)
    monkeypatch.setattr(vertica, "vertica_sql", lambda **kwargs: None)
    executed = vertica.vertica_sql_file(file_path=str(sql_file), conn_id="CID")
    assert executed == 1
    assert conn.close_calls == 1


def test_vertica_sql_dir(monkeypatch, tmp_path):
    (tmp_path / "a.sql").write_text("select 1;", encoding="utf-8")
    (tmp_path / "b.sql").write_text("select 2;select 3;", encoding="utf-8")
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_sql_file", lambda **kwargs: 1 if kwargs["file_path"].endswith("a.sql") else 2)
    out = vertica.vertica_sql_dir(dir_sql=str(tmp_path), conn=conn)
    assert out[str(tmp_path / "a.sql")] == 1
    assert out[str(tmp_path / "b.sql")] == 2
    assert conn.commit_calls == 1


def test_vertica_sql_dir_invalid():
    with pytest.raises(NotADirectoryError):
        vertica.vertica_sql_dir(dir_sql="/no/such/dir", conn=DummyConn())


def test_vertica_select_scalar():
    conn = DummyConn()
    out = vertica.vertica_select_scalar(conn=conn, sql="select 1")
    assert out == 1


def test_vertica_select_scalar_default_when_empty():
    conn = DummyConn()
    conn.cursor_obj._rows = []
    out = vertica.vertica_select_scalar(conn=conn, sql="select 1", default="x")
    assert out == "x"


def test_vertica_insert_dataframe(monkeypatch):
    calls = []
    df = pd.DataFrame([{"id": 1}, {"id": 2}, {"id": 3}])
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_sql", lambda **kwargs: calls.append(kwargs))
    inserted = vertica.vertica_insert_dataframe(
        df=df,
        table_name="tbl",
        conn=conn,
        chunk_size=2,
    )
    assert inserted == 3
    assert len(calls) == 2
    assert conn.commit_calls == 1


def test_vertica_insert_dataframe_empty():
    inserted = vertica.vertica_insert_dataframe(
        df=pd.DataFrame(),
        table_name="tbl",
        conn=DummyConn(),
    )
    assert inserted == 0


def test_vertica_insert_dataframe_validates_chunk_size():
    with pytest.raises(ValueError):
        vertica.vertica_insert_dataframe(
            df=pd.DataFrame([{"id": 1}]),
            table_name="tbl",
            conn=DummyConn(),
            chunk_size=0,
        )


def test_vertica_table_exists_true():
    conn = DummyConn()
    assert vertica.vertica_table_exists(table_name="tbl", conn=conn) is True


def test_vertica_table_exists_false():
    conn = DummyConn()
    conn.cursor_obj._rows = []
    assert vertica.vertica_table_exists(table_name="tbl", conn=conn) is False


def test_vertica_sql_test_uses_test_conn_and_closes(monkeypatch):
    conn = DummyConn()
    monkeypatch.setattr(vertica, "vertica_conn_test", lambda conn_id, dir_path=None, creds_path=None: conn)
    vertica.vertica_sql_test(conn_id="CID", sql="select 1")
    assert conn.close_calls == 1


def test_get_vertica_engine_test(monkeypatch):
    monkeypatch.setattr(
        vertica,
        "un_conn_test",
        lambda conn_id, conn_type, dir_path=None, creds_path=None: ("E", conn_id, conn_type),
    )
    assert vertica.get_vertica_engine_test("CID") == ("E", "CID", "vertica")
